import pathlib
import uuid
from typing import Optional, Dict, Any

import xarray as xr

from AFL.automation.APIServer.Driver import Driver  # type: ignore
from AFL.automation.shared.utilities import mpl_plot_to_bytes,xarray_to_bytes
from AFL.double_agent.Pipeline import Pipeline


class DoubleAgentDriver(Driver):
    """
    Persistent Config
    -----------------
    save_path: str
        path to directory where data will be serialized to
    """

    defaults = {}
    defaults["save_path"] = "/home/AFL/"

    def __init__(
        self,
        name: str = "DoubleAgentDriver",
        overrides: Optional[Dict[str, Any]] = None,
    ):
        Driver.__init__(
            self, name=name, defaults=self.gather_defaults(), overrides=overrides
        )
        self.app = None
        self.name = name

        self.input: Optional[xr.Dataset] = None
        self.pipeline: Optional[Pipeline] = None
        self.results: Dict[str, xr.Dataset] = dict()

    def status(self):
        status = []
        if self.input:
            status.append(f'Input Dims: {self.input.sizes}')
        if self.pipeline:
            status.append(f'Pipeline loaded with {len(self.pipeline.ops)} operations')
        return status
        

    def initialize_input(self, db_uuid: str) -> None:
        """
        Set the initial input data to be evaluated in the `double_agent.Pipeline`

        Parameters
        ----------
        db_uuid: str
            Dropbox UUID to retrieve `xarray.Dataset` from. The Dataset should be deposited using `Client.deposit_obj`
            in interactive mode in order to obtain the uuid of the deposited item. See example below

        Example
        -------
        ```python
        from AFL.automation.APIServer.Client import Client
        import xarray as xr

        client = Client('localhost',port=5053)
        client.login('User')
        db_uuid = client.deposit_obj(obj=xr.Dataset(),interactive=True)['return_val']
        client.enqueue(task_name='initialize_input',db_uuid=db_uuid)
        ```
        """
        self.input = self.retrieve_obj(db_uuid)

    def initialize_pipeline(self, db_uuid: str) -> None:
        """
        Set the `double_agent.Pipeline` to outline

        Parameters
        ----------
        db_uuid: str
            Dropbox UUID to retrieve the `double_agent.Pipeline` from. The Dataset should be deposited using
            `Client.deposit_obj` in interactive mode in order to obtain the uuid of the deposited item. See example
            below

        Example
        -------
        ```python
        from AFL.automation.APIServer.Client import Client
        from AFL.double_agent import *

        client = Client('localhost',port=5053)
        client.login('User')
        db_uuid = client.deposit_obj(obj=Pipeline(),interactive=True)['return_val']
        client.enqueue(task_name='initialize_pipeline',db_uuid=db_uuid)
        ```
        """
        self.pipeline = self.retrieve_obj(db_uuid)

    def append(self, db_uuid: str, concat_dim: str) -> None:
        """

        Parameters
        ----------
        db_uuid: str
            Dropbox UUID to retrieve `xarray.Dataset` from. The Dataset should be deposited using `Client.deposit_obj`
            in interactive mode in order to obtain the uuid of the deposited item. See example below

        concat_dim: str
            `xarray` dimension in input dataset to concatenate to

        """
        if self.input is None:
            raise ValueError(
                'Must set "input" Dataset client.deposit_obj and then DoubleAgentDriver.initialize'
            )

        next_sample = self.retrieve_obj(db_uuid)

        if self.input is None:
            self.input = next_sample
        else:
            self.input = xr.concat(
                [self.input, next_sample], dim=concat_dim, data_vars="minimal"
            )

    @Driver.unqueued(render_hint = 'precomposed_svg')
    def plot_pipeline(self,**kwargs):
        if self.pipeline is not None:
            return mpl_plot_to_bytes(self.pipeline.draw(),format='svg')
        else:
            return None

    @Driver.unqueued(render_hint = 'html')
    def last_result(self,**kwargs):
        return self.last_results._repr_html_()

    @Driver.unqueued(render_hint = 'netcdf')
    def download_last_result(self,**kwargs):
        return xarray_to_bytes(self.last_results)
    
    @Driver.unqueued(render_hint = 'precomposed_png')
    def plot_operation(self,operation,**kwargs):
        try:
            operation = int(operation)
        except ValueError:
            pass
        if self.pipeline is not None:
            if isinstance(operation,str):
                return mpl_plot_to_bytes(self.pipeline.search(operation).plot(),format='png')
            elif isinstance(operation,int):
                return mpl_plot_to_bytes(self.pipeline[operation].plot(),format='png')
            else:
                return None
        else:
            return None
        
    def reset_results(self):
        self.results = dict()

    def predict(
        self,
        deposit: bool = True,
        save_to_disk: bool = True,
        sample_uuid: Optional[str] = None,
        AL_campaign_name: Optional[str] = None,
    ) -> str:
        """
        Evaluate the pipeline set with `.initialize_pipeline`.

        Parameters
        ----------
        deposit: bool
            If True, the `xarray.Dataset` resulting from the `Pipeline` calculation will be placed in this `APIServers`
            dropbox for retrieval and the `db_uuid` will be returned.

        save_to_disk: bool
            If True, the `xarray.Dataset` resulting from the `Pipeline` calculation will be serialized to disk in
            NetCDF format.

        sample_uuid: Optional[str]
            Optionally provide a sample uuid to tag the calculation with

        AL_campaign_name
            Optionally provide an AL campaign name to tag the calculation with

        """
        if (self.pipeline is None) or (self.input is None):
            raise ValueError(
                """Cannot predict without a pipeline and input loaded! Use client.set_driver_object to upload an """
                """Pipeline.Pipeline and an xr.Dataset to this APIServer. You currently have: \n"""
                f"""DoubleAgentDriver.pipeline = {self.pipeline}\n"""
                f"""DoubleAgentDriver.input = {self.input}\n"""
            )
        if sample_uuid is None:
            sample_uuid = 'SAM-'+str(uuid.uuid4())

        ag_uid = "AG-" + str(uuid.uuid4())
        self.results[ag_uid] = self.pipeline.calculate(self.input)

        self.results[ag_uid].attrs['sample_uuid'] = sample_uuid
        self.results[ag_uid].attrs['ag_uuid'] = ag_uid
        self.results[ag_uid].attrs['AL_campaign_name'] = AL_campaign_name

        if save_to_disk:
            path = (
                pathlib.Path(self.config["save_path"])
                / f"{AL_campaign_name}_SAM-{str(sample_uuid)[-6:]}_AG-{ag_uid[-6:]}.nc"
            )
            self.results[ag_uid].to_netcdf(path)

        if deposit:
            self.deposit_obj(self.results[ag_uid], uid=ag_uid)
        
        self.last_results = self.results[ag_uid]
        
        return ag_uid

_OVERRIDE_MAIN_MODULE_NAME = 'DoubleAgentDriver'
if __name__ == '__main__':
    from AFL.automation.shared.launcher import *



"""
Visualization tools for materials composition and phase data.

This module provides plotting utilities for visualizing compositional data and phase
diagrams. It supports both 2D, 3D and ternary plots using matplotlib, with special
handling for labeled data and phase regions.

Key features:
- Support for 2D, 3D and ternary plots
- Surface and scatter plot styles
- Automatic handling of phase labels
- Flexible customization through matplotlib parameters
- Integration with xarray data structures
"""

import itertools
from typing import Union, List, Optional, Dict, Any

import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # Import for 3D plotting
import warnings
from scipy.signal import savgol_filter
import seaborn as sns

from AFL.double_agent.util import listify


def plot_surface_mpl(
        dataset: xr.Dataset,
        component_variable: str,
        component_dim: str = 'component',
        labels: Union[Optional[str], List] = None,
        set_axes_labels: bool = True,
        ternary: bool = False,
        **mpl_kw) -> List[plt.Artist]:
    """Create a surface plot of compositional data using matplotlib.
    
    Generates a filled surface plot of compositional data, supporting both
    2D Cartesian and ternary plots. The surface is colored according to
    phase labels or other specified values.

    Parameters
    ----------
    dataset : xr.Dataset
        Dataset containing compositional data and optional labels

    component_variable : str
        Name of the variable containing component coordinates

    component_dim : str, default='component'
        Name of the dimension containing component names

    labels : Union[Optional[str], List], default=None
        Labels for coloring the surface. Can be:
        - None: Will try to use 'labels' from dataset
        - str: Name of variable in dataset containing labels
        - List: Direct list of label values

    set_axes_labels : bool, default=True
        Whether to set axis labels using component names

    ternary : bool, default=False
        Whether to create a ternary plot for 3-component data.
        If False with 3 components, will create a 3D plot using matplotlib's 3D projection.

    **mpl_kw : dict
        Additional keyword arguments passed to matplotlib's plotting functions

    Returns
    -------
    List[plt.Artist]
        List of created matplotlib artists

    Raises
    ------
    ValueError
        If number of components is not 2 or 3
    ImportError
        If mpltern is not installed for ternary plots
    """
    components = dataset.coords[component_dim]

    if len(components) == 3 and ternary:
        try:
            import mpltern
        except ImportError as e:
            raise ImportError('Could not import mpltern. Please install via conda or pip') from e
        projection = 'ternary'
    elif len(components) == 3 and not ternary:
        projection = '3d'
    elif len(components) == 2:
        projection = None
    else:
        raise ValueError(f'plot_surface only compatible with 2 or 3 components. You passed: {components}')

    coords = dataset[component_variable].transpose(..., component_dim)

    if labels is None:
        if 'labels' in dataset.coords:
            labels = dataset.coords['labels'].values
        elif 'labels' in dataset:
            labels = dataset['labels'].values
        else:
            labels = np.zeros(coords.shape[0])
    elif isinstance(labels, str) and (labels in dataset):
        labels = dataset[labels].values

    if ('cmap' not in mpl_kw) and ('color' not in mpl_kw):
        mpl_kw['cmap'] = 'viridis'
    if 'edgecolor' not in mpl_kw:
        mpl_kw['edgecolor'] = 'face'

    if 'ax' in mpl_kw:
        ax = mpl_kw.pop('ax')
    else:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection=projection)

    # For 3 components with ternary=False, use 3D projection
    if len(components) == 3 and not ternary:
        x, y, z = coords.T
        
        # Create a triangulation for the 3D surface
        from matplotlib.tri import Triangulation
        tri = Triangulation(x, y)
        
        # Plot the triangulated surface in 3D
        artists = ax.plot_trisurf(x, y, z, triangles=tri.triangles, cmap=mpl_kw.get('cmap'), 
                                 edgecolor=mpl_kw.get('edgecolor'), alpha=mpl_kw.get('alpha', 0.8))
        
        # Add color based on labels if provided
        if labels is not None:
            artists.set_array(labels)
    else:
        artists = ax.tripcolor(*coords.T, labels, **mpl_kw)

    if set_axes_labels:
        if projection == 'ternary':
            labels_dict = {k: v.values[()] for k, v in zip(['tlabel', 'llabel', 'rlabel'], components)}
            ax.set(**labels_dict)
        elif projection == '3d':
            ax.set_xlabel(components[0].values[()])
            ax.set_ylabel(components[1].values[()])
            ax.set_zlabel(components[2].values[()])
        else:
            labels_dict = {k: v.values[()] for k, v in zip(['xlabel', 'ylabel'], components)}
            ax.set(**labels_dict)
        
        if projection != '3d':  # Only set grid for 2D plots
            ax.grid('on', color='black')
    
    return artists


def plot_scatter_mpl(
        dataset: xr.Dataset,
        component_variable: str,
        component_dim: str = 'component',
        labels: Union[Optional[str], List] = None,
        discrete_labels: bool = True,
        set_axes_labels: bool = True,
        ternary: bool = False,
        **mpl_kw) -> List[plt.Artist]:
    """Create a scatter plot of compositional data using matplotlib.
    
    Generates a scatter plot of compositional data, supporting both 2D
    Cartesian and ternary plots. Points can be colored and marked according
    to phase labels or other values.

    Parameters
    ----------
    dataset : xr.Dataset
        Dataset containing compositional data and optional labels

    component_variable : str
        Name of the variable containing component coordinates

    component_dim : str, default='component'
        Name of the dimension containing component names

    labels : Union[Optional[str], List], default=None
        Labels for coloring points. Can be:
        - None: Will try to use 'labels' from dataset
        - str: Name of variable in dataset containing labels
        - List: Direct list of label values

    discrete_labels : bool, default=True
        Whether to treat labels as discrete categories (True) or continuous values (False)

    set_axes_labels : bool, default=True
        Whether to set axis labels using component names

    ternary : bool, default=False
        Whether to create a ternary plot for 3-component data.
        If False with 3 components, will create a 3D plot using matplotlib's 3D projection.

    **mpl_kw : dict
        Additional keyword arguments passed to matplotlib's scatter

    Returns
    -------
    List[plt.Artist]
        List of created matplotlib artists

    Raises
    ------
    ValueError
        If number of components is not 2 or 3
    ImportError
        If mpltern is not installed for ternary plots

    Notes
    -----
    When discrete_labels is True, different marker styles are used for each
    unique label value, cycling through a predefined set of markers.
    """
    components = dataset.coords[component_dim]

    if len(components) == 3 and ternary:
        try:
            import mpltern
        except ImportError as e:
            raise ImportError('Could not import mpltern. Please install via conda or pip') from e
        projection = 'ternary'
    elif len(components) == 3 and not ternary:
        projection = '3d'
    elif len(components) == 2:
        projection = None
    else:
        raise ValueError(f'plot_surface only compatible with 2 or 3 components. You passed: {components}')

    coords = dataset[component_variable].transpose(..., component_dim).values

    if (labels is None):
        if ('labels' in dataset.coords):
            labels = dataset.coords['labels'].values
        elif ('labels' in dataset):
            labels = dataset['labels'].values
        else:
            labels = np.zeros(coords.shape[0])
    elif isinstance(labels, str) and (labels in dataset):
        labels = dataset[labels].values

    if 'ax' in mpl_kw:
        ax = mpl_kw.pop('ax')
    else:
        fig = plt.figure()
        ax = fig.add_subplot(111, projection=projection)

    # For 3 components with ternary=False, use 3D projection
    if len(components) == 3 and not ternary:
        x, y, z = coords.T
        
        if discrete_labels:
            markers = itertools.cycle(listify(mpl_kw.get('marker', ['^', 'v', '<', '>', 'o', 'd', 'p', 'x'])))
            artists = []
            for label in np.unique(labels):
                mask = (labels == label)
                mpl_kw['marker'] = next(markers)
                # Use scatter3D for 3D scatter plots
                artists.append(ax.scatter3D(x[mask], y[mask], z[mask], **mpl_kw))
        else:
            # Use scatter3D for 3D scatter plots
            artists = ax.scatter3D(x, y, z, c=labels, **mpl_kw)
    else:
        if discrete_labels:
            markers = itertools.cycle(listify(mpl_kw.get('marker', ['^', 'v', '<', '>', 'o', 'd', 'p', 'x'])))
            artists = []
            for label in np.unique(labels):
                mask = (labels == label)
                mpl_kw['marker'] = next(markers)
                artists.append(ax.scatter(*coords[mask].T, **mpl_kw))
        else:
            artists = ax.scatter(*coords.T, c=labels, **mpl_kw)

    if set_axes_labels:
        if projection == 'ternary':
            labels_dict = {k: v.values[()] for k, v in zip(['tlabel', 'llabel', 'rlabel'], components)}
            ax.set(**labels_dict)
        elif projection == '3d':
            ax.set_xlabel(components[0].values[()])
            ax.set_ylabel(components[1].values[()])
            ax.set_zlabel(components[2].values[()])
        else:
            labels_dict = {k: v.values[()] for k, v in zip(['xlabel', 'ylabel'], components)}
            ax.set(**labels_dict)
        
        if projection != '3d':  # Only set grid for 2D plots
            ax.grid('on', color='black')
    
    return artists


def plot_surface_plotly(
        dataset: xr.Dataset,
        component_variable: str,
        component_dim: str = 'component',
        labels: Union[Optional[str], List] = None,
        set_axes_labels: bool = True,
        ternary: bool = False,
        **plotly_kw) -> Dict[str, Any]:
    """Create a surface plot of compositional data using Plotly.
    
    Generates an interactive filled surface plot of compositional data, supporting both
    2D Cartesian, 3D, and ternary plots. The surface is colored according to
    phase labels or other specified values.

    Parameters
    ----------
    dataset : xr.Dataset
        Dataset containing compositional data and optional labels

    component_variable : str
        Name of the variable containing component coordinates

    component_dim : str, default='component'
        Name of the dimension containing component names

    labels : Union[Optional[str], List], default=None
        Labels for coloring the surface. Can be:
        - None: Will try to use 'labels' from dataset
        - str: Name of variable in dataset containing labels
        - List: Direct list of label values

    set_axes_labels : bool, default=True
        Whether to set axis labels using component names

    ternary : bool, default=False
        Whether to create a ternary plot for 3-component data.
        If False with 3 components, will create a 3D plot.

    **plotly_kw : dict
        Additional keyword arguments passed to plotly's plotting functions

    Returns
    -------
    Dict[str, Any]
        Plotly figure object

    Raises
    ------
    ValueError
        If number of components is not 2 or 3
    ImportError
        If plotly is not installed
    """
    try:
        import plotly.graph_objects as go
        from plotly.subplots import make_subplots
    except ImportError as e:
        raise ImportError('Could not import plotly. Please install via pip install plotly') from e

    components = dataset.coords[component_dim]

    if len(components) not in [2, 3]:
        raise ValueError(f'plot_surface only compatible with 2 or 3 components. You passed: {components}')

    coords = dataset[component_variable].transpose(..., component_dim)

    if labels is None:
        if 'labels' in dataset.coords:
            labels = dataset.coords['labels'].values
        elif 'labels' in dataset:
            labels = dataset['labels'].values
        else:
            labels = np.zeros(coords.shape[0])
    elif isinstance(labels, str) and (labels in dataset):
        labels = dataset[labels].values

    # Set default colormap if not provided
    colorscale = plotly_kw.get('colorscale', 'Viridis')
    
    # Set default figure size (similar to matplotlib's default size of 6.4 x 4.8 inches)
    # Plotly uses pixels, so convert inches to pixels (assuming 100 pixels per inch)
    width = plotly_kw.get('width', 640)  # 6.4 inches * 100 pixels/inch
    height = plotly_kw.get('height', 480)  # 4.8 inches * 100 pixels/inch
    
    # Create figure with appropriate layout
    if len(components) == 3 and ternary:
        fig = go.Figure()
        
        # Extract coordinates
        a, b, c = coords.T
        
        # Create ternary scatter plot with color based on labels
        fig.add_trace(go.Scatterternary(
            a=a, b=b, c=c,
            mode='markers',
            marker=dict(
                color=labels,
                colorscale=colorscale,
                size=10,
                showscale=True
            ),
            hovertext=[f'Label: {l}' for l in labels],
        ))
        
        # Set layout for ternary plot
        fig.update_layout(
            width=width,
            height=height,
            ternary=dict(
                aaxis=dict(title=components[0].values[()], min=0, linewidth=2),
                baxis=dict(title=components[1].values[()], min=0, linewidth=2),
                caxis=dict(title=components[2].values[()], min=0, linewidth=2),
                bgcolor='white'
            ),
            **plotly_kw.get('layout', {})
        )
    
    elif len(components) == 3 and not ternary:
        fig = go.Figure()
        
        # Extract coordinates
        x, y, z = coords.T
        
        # Create a mesh3d surface
        from scipy.spatial import Delaunay
        tri = Delaunay(np.column_stack([x, y]))
        
        fig.add_trace(go.Mesh3d(
            x=x, y=y, z=z,
            i=tri.simplices[:, 0],
            j=tri.simplices[:, 1],
            k=tri.simplices[:, 2],
            intensity=labels,
            colorscale=colorscale,
            opacity=0.8,
            **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'colorscale', 'width', 'height']}
        ))
        
        # Set layout for 3D plot
        layout_dict = {'width': width, 'height': height}
        if set_axes_labels:
            layout_dict['scene'] = dict(
                xaxis_title=components[0].values[()],
                yaxis_title=components[1].values[()],
                zaxis_title=components[2].values[()]
            )
        
        fig.update_layout(**layout_dict, **plotly_kw.get('layout', {}))
    
    else:  # 2D plot
        fig = go.Figure()
        
        # Extract coordinates
        x, y = coords.T
        
        # Create a contour plot for 2D surface
        from scipy.interpolate import griddata
        
        # Create a grid for interpolation
        xi = np.linspace(min(x), max(x), 100)
        yi = np.linspace(min(y), max(y), 100)
        xi_grid, yi_grid = np.meshgrid(xi, yi)
        
        # Interpolate labels onto the grid
        zi = griddata((x, y), labels, (xi_grid, yi_grid), method='linear')
        
        # Create contour plot
        fig.add_trace(go.Contour(
            z=zi,
            x=xi,
            y=yi,
            colorscale=colorscale,
            **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'colorscale', 'width', 'height']}
        ))
        
        # Set layout for 2D plot
        layout_dict = {'width': width, 'height': height}
        if set_axes_labels:
            layout_dict['xaxis_title'] = components[0].values[()]
            layout_dict['yaxis_title'] = components[1].values[()]
        
        fig.update_layout(**layout_dict, **plotly_kw.get('layout', {}))
    
    return fig


def plot_scatter_plotly(
        dataset: xr.Dataset,
        component_variable: str,
        component_dim: str = 'component',
        labels: Union[Optional[str], List] = None,
        discrete_labels: bool = True,
        set_axes_labels: bool = True,
        ternary: bool = False,
        **plotly_kw) -> Dict[str, Any]:
    """Create a scatter plot of compositional data using Plotly.
    
    Generates an interactive scatter plot of compositional data, supporting both 2D
    Cartesian, 3D, and ternary plots. Points can be colored and marked according
    to phase labels or other values.

    Parameters
    ----------
    dataset : xr.Dataset
        Dataset containing compositional data and optional labels

    component_variable : str
        Name of the variable containing component coordinates

    component_dim : str, default='component'
        Name of the dimension containing component names

    labels : Union[Optional[str], List], default=None
        Labels for coloring points. Can be:
        - None: Will try to use 'labels' from dataset
        - str: Name of variable in dataset containing labels
        - List: Direct list of label values

    discrete_labels : bool, default=True
        Whether to treat labels as discrete categories (True) or continuous values (False)

    set_axes_labels : bool, default=True
        Whether to set axis labels using component names

    ternary : bool, default=False
        Whether to create a ternary plot for 3-component data.
        If False with 3 components, will create a 3D plot.

    **plotly_kw : dict
        Additional keyword arguments passed to plotly's plotting functions

    Returns
    -------
    Dict[str, Any]
        Plotly figure object

    Raises
    ------
    ValueError
        If number of components is not 2 or 3
    ImportError
        If plotly is not installed

    Notes
    -----
    When discrete_labels is True, different marker symbols are used for each
    unique label value, cycling through a predefined set of markers.
    """
    try:
        import plotly.graph_objects as go
        from plotly.subplots import make_subplots
    except ImportError as e:
        raise ImportError('Could not import plotly. Please install via pip install plotly') from e

    components = dataset.coords[component_dim]

    if len(components) not in [2, 3]:
        raise ValueError(f'plot_scatter only compatible with 2 or 3 components. You passed: {components}')

    coords = dataset[component_variable].transpose(..., component_dim).values

    if (labels is None):
        if ('labels' in dataset.coords):
            labels = dataset.coords['labels'].values
        elif ('labels' in dataset):
            labels = dataset['labels'].values
        else:
            labels = np.zeros(coords.shape[0])
    elif isinstance(labels, str) and (labels in dataset):
        labels = dataset[labels].values

    # Set default colormap if not provided
    colorscale = plotly_kw.get('colorscale', 'Viridis')
    
    # Set default figure size (similar to matplotlib's default size of 6.4 x 4.8 inches)
    # Plotly uses pixels, so convert inches to pixels (assuming 100 pixels per inch)
    width = plotly_kw.get('width', 640)  # 6.4 inches * 100 pixels/inch
    height = plotly_kw.get('height', 480)  # 4.8 inches * 100 pixels/inch
    
    # Create figure
    fig = go.Figure()
    
    # Define marker symbols for discrete labels
    plotly_symbols = ['circle', 'square', 'diamond', 'cross', 'x', 'triangle-up', 
                      'triangle-down', 'triangle-left', 'triangle-right', 'pentagon', 
                      'hexagon', 'star', 'hexagram', 'star-triangle-up', 'star-triangle-down']
    
    if len(components) == 3 and ternary:
        # Extract coordinates
        a, b, c = coords.T
        
        if discrete_labels:
            # Create a trace for each unique label
            for i, label in enumerate(np.unique(labels)):
                mask = (labels == label)
                symbol = plotly_symbols[i % len(plotly_symbols)]
                
                fig.add_trace(go.Scatterternary(
                    a=a[mask], b=b[mask], c=c[mask],
                    mode='markers',
                    marker=dict(
                        symbol=symbol,
                        size=10,
                        color=i,  # Use index for color
                    ),
                    name=f'Label {label}',
                    **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'colorscale', 'width', 'height']}
                ))
        else:
            # Create a single trace with continuous color scale
            fig.add_trace(go.Scatterternary(
                a=a, b=b, c=c,
                mode='markers',
                marker=dict(
                    color=labels,
                    colorscale=colorscale,
                    size=10,
                    showscale=True
                ),
                hovertext=[f'Label: {l}' for l in labels],
                **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'colorscale', 'width', 'height']}
            ))
        
        # Set layout for ternary plot
        fig.update_layout(
            width=width,
            height=height,
            ternary=dict(
                aaxis=dict(title=components[0].values[()], min=0, linewidth=2),
                baxis=dict(title=components[1].values[()], min=0, linewidth=2),
                caxis=dict(title=components[2].values[()], min=0, linewidth=2),
                bgcolor='white'
            ),
            **plotly_kw.get('layout', {})
        )
    
    elif len(components) == 3 and not ternary:
        # Extract coordinates
        x, y, z = coords.T
        
        if discrete_labels:
            # Create a trace for each unique label
            for i, label in enumerate(np.unique(labels)):
                mask = (labels == label)
                symbol = plotly_symbols[i % len(plotly_symbols)]
                
                fig.add_trace(go.Scatter3d(
                    x=x[mask], y=y[mask], z=z[mask],
                    mode='markers',
                    marker=dict(
                        symbol=symbol,
                        size=6,
                        color=i,  # Use index for color
                    ),
                    name=f'Label {label}',
                    **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'colorscale', 'width', 'height']}
                ))
        else:
            # Create a single trace with continuous color scale
            fig.add_trace(go.Scatter3d(
                x=x, y=y, z=z,
                mode='markers',
                marker=dict(
                    color=labels,
                    colorscale=colorscale,
                    size=6,
                    showscale=True
                ),
                hovertext=[f'Label: {l}' for l in labels],
                **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'colorscale', 'width', 'height']}
            ))
        
        # Set layout for 3D plot
        layout_dict = {'width': width, 'height': height}
        if set_axes_labels:
            layout_dict['scene'] = dict(
                xaxis_title=components[0].values[()],
                yaxis_title=components[1].values[()],
                zaxis_title=components[2].values[()]
            )
        
        fig.update_layout(**layout_dict, **plotly_kw.get('layout', {}))
    
    else:  # 2D plot
        # Extract coordinates
        x, y = coords.T
        
        if discrete_labels:
            # Create a trace for each unique label
            for i, label in enumerate(np.unique(labels)):
                mask = (labels == label)
                symbol = plotly_symbols[i % len(plotly_symbols)]
                
                fig.add_trace(go.Scatter(
                    x=x[mask], y=y[mask],
                    mode='markers',
                    marker=dict(
                        symbol=symbol,
                        size=10,
                        color=i,  # Use index for color
                    ),
                    name=f'Label {label}',
                    **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'colorscale', 'width', 'height']}
                ))
        else:
            # Create a single trace with continuous color scale
            fig.add_trace(go.Scatter(
                x=x, y=y,
                mode='markers',
                marker=dict(
                    color=labels,
                    colorscale=colorscale,
                    size=10,
                    showscale=True
                ),
                hovertext=[f'Label: {l}' for l in labels],
                **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'colorscale', 'width', 'height']}
            ))
        
        # Set layout for 2D plot
        layout_dict = {'width': width, 'height': height}
        if set_axes_labels:
            layout_dict['xaxis_title'] = components[0].values[()]
            layout_dict['yaxis_title'] = components[1].values[()]
        
        fig.update_layout(**layout_dict, **plotly_kw.get('layout', {}))
    
    return fig


def plot_sas_mpl(
        dataset: xr.Dataset,
        plot_type: str = 'loglog',
        x: str = 'q',
        y: str = None,
        ylabel: str = 'Intensity [A.U.]',
        xlabel: str = r'q [$Å^{-1}$]',
        legend: bool = True,
        base: float = 10,
        waterfall: bool = False,
        clean_params: dict = None,
        **mpl_kw) -> List[plt.Artist]:
    """Create a plot of small-angle scattering data using matplotlib.
    
    Generates a plot of scattering data with various options for visualization,
    including log-log plots, waterfalls.

    Parameters
    ----------
    dataset : xr.Dataset
        Dataset containing scattering data

    plot_type : str, default='loglog'
        Type of plot to create. Options:
        - 'loglog': Log-log plot
        - 'linlin': Linear-linear plot
        - 'logwaterfall': Log waterfall plot
        - 'waterfall': Linear waterfall plot

    x : str, default='q'
        Name of the x-axis variable

    y : str, default=None
        Name of the y-axis variable. If None, will use the first data variable that isn't x.

    ylabel : str, default='Intensity [A.U.]'
        Label for the y-axis

    xlabel : str, default='q [$Å^{-1}$]'
        Label for the x-axis

    legend : bool, default=True
        Whether to show a legend

    base : float, default=10
        Base for waterfall plots

    waterfall : bool, default=False
        Whether to create a waterfall plot (overrides plot_type)

    **mpl_kw : dict
        Additional keyword arguments passed to matplotlib's plotting functions

    Returns
    -------
    List[plt.Artist]
        List of created matplotlib artists
    """
    # Find the y variable if not specified
    if y is None:
        for var in dataset.data_vars:
            if var != x:
                y = var
                break
        if y is None:
            raise ValueError("Could not find a data variable to plot. Please specify 'y'.")
    
    # Create a temporary DataArray for the y variable
    temp_da = dataset[y].copy()
    
    # Determine which plotting method to use and create the plot
    if waterfall:
        if plot_type == 'loglog' or plot_type == 'logwaterfall':
            return _plot_logwaterfall(temp_da, x=x, ylabel=ylabel, xlabel=xlabel, legend=legend, base=base, **mpl_kw)
        else:
            return _plot_waterfall(temp_da, x=x, ylabel=ylabel, xlabel=xlabel, legend=legend, base=base, **mpl_kw)
    elif plot_type == 'loglog':
        return _plot_loglog(temp_da, x=x, ylabel=ylabel, xlabel=xlabel, legend=legend, **mpl_kw)
    elif plot_type == 'linlin':
        return _plot_linlin(temp_da, x=x, ylabel=ylabel, xlabel=xlabel, legend=legend, **mpl_kw)
    elif plot_type == 'logwaterfall':
        return _plot_logwaterfall(temp_da, x=x, ylabel=ylabel, xlabel=xlabel, legend=legend, base=base, **mpl_kw)
    elif plot_type == 'waterfall':
        return _plot_waterfall(temp_da, x=x, ylabel=ylabel, xlabel=xlabel, legend=legend, base=base, **mpl_kw)
    else:
        raise ValueError(f"Unknown plot_type: {plot_type}. Options are 'loglog', 'linlin', 'logwaterfall', 'waterfall'")


def plot_sas_plotly(
        dataset: xr.Dataset,
        plot_type: str = 'loglog',
        x: str = 'q',
        y: str = None,
        ylabel: str = 'Intensity [A.U.]',
        xlabel: str = r'q [$Å^{-1}$]',
        legend: bool = True,
        base: float = 10,
        waterfall: bool = False,
        **plotly_kw) -> Dict[str, Any]:
    """Create a plot of small-angle scattering data using Plotly.
    
    Generates an interactive plot of scattering data with various options for visualization,
    including log-log plots, waterfalls, and data cleaning.

    Parameters
    ----------
    dataset : xr.Dataset
        Dataset containing scattering data

    plot_type : str, default='loglog'
        Type of plot to create. Options:
        - 'loglog': Log-log plot
        - 'linlin': Linear-linear plot
        - 'logwaterfall': Log waterfall plot
        - 'waterfall': Linear waterfall plot

    x : str, default='q'
        Name of the x-axis variable

    y : str, default=None
        Name of the y-axis variable. If None, will use the first data variable that isn't x.

    ylabel : str, default='Intensity [A.U.]'
        Label for the y-axis

    xlabel : str, default='q [$Å^{-1}$]'
        Label for the x-axis

    legend : bool, default=True
        Whether to show a legend

    base : float, default=10
        Base for waterfall plots

    waterfall : bool, default=False
        Whether to create a waterfall plot (overrides plot_type)

    **plotly_kw : dict
    clean_params : dict, default=None
        Parameters for data cleaning. Options include:
        - qlo: Lower q limit
        - qhi: Upper q limit
        - qlo_isel: Lower q index
        - qhi_isel: Upper q index
        - pedestal: Pedestal value
        - npts: Number of points
        - derivative: Derivative order
        - sgf_window_length: Savitzky-Golay filter window length
        - sgf_polyorder: Savitzky-Golay filter polynomial order
        - apply_log_scale: Whether to apply log scale

    **plotly_kw : dict
        Additional keyword arguments passed to plotly's plotting functions

    Returns
    -------
    Dict[str, Any]
        Plotly figure object
    """
    try:
        import plotly.graph_objects as go
    except ImportError as e:
        raise ImportError('Could not import plotly. Please install via pip install plotly') from e
    
    # Find the y variable if not specified
    if y is None:
        for var in dataset.data_vars:
            if var != x:
                y = var
                break
        if y is None:
            raise ValueError("Could not find a data variable to plot. Please specify 'y'.")
    
    # Create a temporary DataArray for the y variable
    temp_da = dataset[y].copy()
    
    # Set default figure size (similar to matplotlib's default size of 6.4 x 4.8 inches)
    # Plotly uses pixels, so convert inches to pixels (assuming 100 pixels per inch)
    width = plotly_kw.get('width', 640)  # 6.4 inches * 100 pixels/inch
    height = plotly_kw.get('height', 480)  # 4.8 inches * 100 pixels/inch
    
    # Create figure
    fig = go.Figure()
    
    # Get the dimension to iterate over (usually 'sample')
    dim = next((d for d in temp_da.dims if d != x), None)
    
    # Determine plot type and create appropriate traces
    if waterfall or plot_type in ['logwaterfall', 'waterfall']:
        # Calculate multiplier for waterfall effect
        N = temp_da[dim].shape[0] if dim else 1
        
        if plot_type == 'loglog' or plot_type == 'logwaterfall' or (waterfall and plot_type == 'loglog'):
            # Log waterfall
            mul = np.geomspace(1, float(base) ** (N + 1), N) if dim else [1]
            for i, m in enumerate(mul):
                if dim:
                    y_data = temp_da.isel({dim: i}).values * m
                    name = str(temp_da[dim].isel({dim: i}).values)
                else:
                    y_data = temp_da.values * m
                    name = 'data'
                
                fig.add_trace(go.Scatter(
                    x=temp_da[x].values,
                    y=y_data,
                    mode='markers',
                    name=name,
                    **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'width', 'height']}
                ))
            
            # Set log scales
            fig.update_layout(
                xaxis_type="log",
                yaxis_type="log"
            )
        else:
            # Linear waterfall
            mul = np.linspace(1, base * N, N) if dim else [1]
            for i, m in enumerate(mul):
                if dim:
                    y_data = temp_da.isel({dim: i}).values + m
                    name = str(temp_da[dim].isel({dim: i}).values)
                else:
                    y_data = temp_da.values + m
                    name = 'data'
                
                fig.add_trace(go.Scatter(
                    x=temp_da[x].values,
                    y=y_data,
                    mode='markers',
                    name=name,
                    **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'width', 'height']}
                ))
    else:
        # Regular plots (loglog or linlin)
        if dim:
            for i in range(temp_da[dim].shape[0]):
                y_data = temp_da.isel({dim: i}).values
                name = str(temp_da[dim].isel({dim: i}).values)
                
                fig.add_trace(go.Scatter(
                    x=temp_da[x].values,
                    y=y_data,
                    mode='markers',
                    name=name,
                    **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'width', 'height']}
                ))
        else:
            fig.add_trace(go.Scatter(
                x=temp_da[x].values,
                y=temp_da.values,
                mode='markers',
                name='data',
                **{k: v for k, v in plotly_kw.items() if k not in ['layout', 'width', 'height']}
            ))
        
        # Set scales based on plot type
        if plot_type == 'loglog':
            fig.update_layout(
                xaxis_type="log",
                yaxis_type="log"
            )
    
    # Set layout
    layout_dict = {
        'width': width,
        'height': height,
        'xaxis_title': xlabel,
        'yaxis_title': ylabel,
        'showlegend': legend
    }
    
    # Update layout with user-provided settings
    fig.update_layout(**layout_dict, **plotly_kw.get('layout', {}))
    
    return fig


def _plot_logwaterfall(data: xr.DataArray, x: str = 'q', dim: str = None, ylabel: str = 'Intensity [A.U.]', 
                      xlabel: str = r'q [$Å^{-1}$]', legend: bool = True, base: float = 10, ax=None, **mpl_kw):
    """Create a log waterfall plot of scattering data."""
    if dim is None:
        dim = next((d for d in data.dims if d != x), None)
    
    if dim is None:
        # Single curve, no waterfall effect needed
        return _plot_loglog(data, x=x, ylabel=ylabel, xlabel=xlabel, legend=legend, ax=ax, **mpl_kw)
    
    N = data[dim].shape[0]
    mul = data[dim].copy(data=np.geomspace(1, float(base) ** (N + 1), N))
    
    lines = (
        data
        .pipe(lambda x: x * mul)
        .plot
        .line(
            x=x,
            xscale='log',
            yscale='log',
            marker='.',
            ls='None',
            add_legend=legend,
            ax=ax,
            **mpl_kw)
    )
    
    plt.gca().set(xlabel=xlabel, ylabel=ylabel)
    if legend and (len(lines) > 1):
        sns.move_legend(plt.gca(), loc=6, bbox_to_anchor=(1.05, 0.5))
    
    return lines


def _plot_waterfall(data: xr.DataArray, x: str = 'q', dim: str = None, ylabel: str = 'Intensity [A.U.]', 
                   xlabel: str = r'q [$Å^{-1}$]', legend: bool = True, base: float = 1, ax=None, **mpl_kw):
    """Create a linear waterfall plot of scattering data."""
    if dim is None:
        dim = next((d for d in data.dims if d != x), None)
    
    if dim is None:
        # Single curve, no waterfall effect needed
        return _plot_linlin(data, x=x, ylabel=ylabel, xlabel=xlabel, legend=legend, ax=ax, **mpl_kw)
    
    N = data[dim].shape[0]
    mul = data[dim].copy(data=np.linspace(1, base * N, N))
    
    lines = (
        data
        .pipe(lambda x: x + mul)
        .plot
        .line(
            x=x,
            xscale='linear',
            yscale='linear',
            marker='.',
            ls='None',
            add_legend=legend,
            ax=ax,
            **mpl_kw)
    )
    
    plt.gca().set(xlabel=xlabel, ylabel=ylabel)
    if legend and (len(lines) > 1):
        sns.move_legend(plt.gca(), loc=6, bbox_to_anchor=(1.05, 0.5))
    
    return lines


def _plot_loglog(data: xr.DataArray, x: str = 'q', ylabel: str = 'Intensity [A.U.]', 
                xlabel: str = r'q [$Å^{-1}$]', legend: bool = True, ax=None, **mpl_kw):
    """Create a log-log plot of scattering data."""
    lines = data.plot.line(
        x=x,
        xscale='log',
        yscale='log',
        marker='.',
        ls='None',
        add_legend=legend,
        ax=ax,
        **mpl_kw)
    
    plt.gca().set(xlabel=xlabel, ylabel=ylabel)
    if legend and (len(lines) > 1):
        sns.move_legend(plt.gca(), loc=6, bbox_to_anchor=(1.05, 0.5))
    
    return lines


def _plot_linlin(data: xr.DataArray, x: str = 'q', ylabel: str = 'Intensity [A.U.]', 
                xlabel: str = r'q [$Å^{-1}$]', legend: bool = True, ax=None, **mpl_kw):
    """Create a linear-linear plot of scattering data."""
    lines = data.plot.line(
        x=x,
        marker='.',
        ls='None',
        add_legend=legend,
        ax=ax,
        **mpl_kw)
    
    plt.gca().set(xlabel=xlabel, ylabel=ylabel)
    if legend and (len(lines) > 1):
        sns.move_legend(plt.gca(), loc=6, bbox_to_anchor=(1.05, 0.5))
    
    return lines


def plot(
        dataset: xr.Dataset,
        kind: str = 'scatter',
        backend: str = 'mpl',
        **kwargs) -> Union[List[plt.Artist], Dict[str, Any]]:
    """Master plotting function that dispatches to specialized plotting functions.
    
    This function serves as a unified interface for all plotting functions in the module.
    It dispatches to the appropriate specialized function based on the plottype and backend.
    
    Parameters
    ----------
    dataset : xr.Dataset
        Dataset containing the data to plot
    
    plottype : str, default='scatter'
        Type of plot to create. Options:
        - 'scatter': Scatter plot of compositional data
        - 'surface': Surface plot of compositional data
        - 'sas': Small-angle scattering plot
    
    backend : str, default='mpl'
        Plotting backend to use. Options:
        - 'mpl': Matplotlib
        - 'plotly': Plotly
    
    **kwargs : dict
        Additional keyword arguments passed to the specialized plotting function.
        See the documentation of the specialized functions for details.
    
    Returns
    -------
    Union[List[plt.Artist], Dict[str, Any]]
        The return value of the specialized plotting function, which is either
        a list of matplotlib artists or a plotly figure object.
    
    Raises
    ------
    ValueError
        If an invalid plottype or backend is specified.
    
    Examples
    --------
    >>> # Create a scatter plot using matplotlib
    >>> plot(dataset, plottype='scatter', backend='mpl', 
    ...      component_variable='compositions', labels='labels')
    
    >>> # Create a surface plot using plotly
    >>> plot(dataset, plottype='surface', backend='plotly', 
    ...      component_variable='compositions', ternary=True)
    
    >>> # Create a small-angle scattering plot using matplotlib
    >>> plot(dataset, plottype='sas', backend='mpl', 
    ...      plot_type='loglog', x='q', y='intensity')
    """
    # Validate plottype
    valid_plottypes = ['scatter', 'surface', 'sas']
    if kind not in valid_plottypes:
        raise ValueError(f"Invalid plottype: {kind}. Must be one of {valid_plottypes}")
    
    # Validate backend
    valid_backends = ['mpl', 'plotly']
    if backend not in valid_backends:
        raise ValueError(f"Invalid backend: {backend}. Must be one of {valid_backends}")
    
    # Dispatch to the appropriate specialized function
    if kind == 'scatter':
        if backend == 'mpl':
            return plot_scatter_mpl(dataset, **kwargs)
        else:  # backend == 'plotly'
            return plot_scatter_plotly(dataset, **kwargs)
    
    elif kind == 'surface':
        if backend == 'mpl':
            return plot_surface_mpl(dataset, **kwargs)
        else:  # backend == 'plotly'
            return plot_surface_plotly(dataset, **kwargs)
    
    else:  # kind == 'sas'
        if backend == 'mpl':
            return plot_sas_mpl(dataset, **kwargs)
        else:  # backend == 'plotly'
            return plot_sas_plotly(dataset, **kwargs)



from AFL.automation.APIServer.Client import Client
from AFL.automation.prepare.OT2Client import OT2Client
from AFL.automation.shared.utilities import listify
from AFL.automation.APIServer.Driver import Driver
from AFL.agent.AgentClient import AgentClient
from AFL.automation.shared.units import units

import warnings

import pandas as pd
import numpy as np 
import xarray as xr 
from math import ceil,sqrt
import json
import time
import requests
import shutil
import datetime
import traceback
import pathlib
import uuid
import copy
import os

import AFL.automation.prepare
import shutil
import h5py

from tiled.client import from_uri
from tiled.queries import Eq,Contains
from scipy.stats import iqr

class Multimodal_AL_SampleDriver(Driver):
    defaults={}
    defaults['snapshot_directory'] = '/home/nistoroboto'
    defaults['csv_data_path'] = '/nfs/aux/chess/reduced_data/cycles/2022-1/id3b/beaucage-2324-D/analysis/'
    defaults['AL_manifest_file'] = '/nfs/aux/chess/reduced_data/cycles/2022-1/id3b/beaucage-2324-D/analysis/data_manifest.csv'
    defaults['data_tag'] = 'default'
    defaults['max_sample_transmission'] = 0.6
    defaults['spec_data_path'] = '/home/pi/2305_SINQ_data_reduced/'
    def __init__(self,
            load_url,
            prep_url,
            sas_url,
            agent_url,
            tiled_url,
            turb_url,
            spec_url=None,
            camera_urls = None,
            snapshot_directory =None,
            overrides=None, 
            ):

        Driver.__init__(self,name='Multimodal_AL_SampleDriver',defaults=self.gather_defaults(),overrides=overrides)

        if not (len(load_url.split(':'))==2):
            raise ArgumentError('Need to specify both ip and port on load_url')

        if not (len(prep_url.split(':'))==2):
            raise ArgumentError('Need to specify both ip and port on prep_url')

        if not (len(agent_url.split(':'))==2):
            raise ArgumentError('Need to specify both ip and port on agent_url')

        self.app = None
        self.name = 'Multimodal_AL_SampleDriver'

        #prepare samples
        self.prep_url = prep_url
        self.prep_client = OT2Client(prep_url.split(':')[0],port=prep_url.split(':')[1])
        self.prep_client.login('SampleServer_PrepClient')
        self.prep_client.debug(False)

        #load samples
        self.load_client = Client(load_url.split(':')[0],port=load_url.split(':')[1])
        self.load_client.login('SampleServer_LoadClient')
        self.load_client.debug(False)
# 
        #measure samples
        self.sas_url = sas_url
        self.sas_client = Client(sas_url.split(':')[0],port=sas_url.split(':')[1])
        self.sas_client.login('SampleServer_SASClient')
        self.sas_client.debug(False)
        
        #agent
        self.agent_url = agent_url
        self.agent_client = AgentClient(agent_url.split(':')[0],port=agent_url.split(':')[1])
        self.agent_client.login('SampleServer_AgentClient')
        self.agent_client.debug(False)

        # start tiled catalog connection
        self.tiled_cat = from_uri(tiled_url,api_key=os.environ['TILED_API_KEY'])

        if spec_url is not None:
            self.spec_url = spec_url
            self.spec_client = Client(spec_url.split(':')[0],port=spec_url.split(':')[1])
            self.spec_client.login('SampleServer_LSClient')
            self.spec_client.debug(False)
            self.spec_client.enqueue(task_name='setExposure',time=0.1)
        else:
            self.spec_client = None

        self.turb_url = turb_url
        self.turb_client = Client(turb_url.split(':')[0],port=turb_url.split(':')[1])
        self.turb_client.login('SampleServer_TurbClient')
        self.turb_client.debug(False)

        self.camera_urls = camera_urls

        if snapshot_directory is not None:
            self.config['snapshot_directory'] = snapshot_directory

        self.rinse_uuid = None
        self.prep_uuid = None
        self.catch_uuid = None
        self.agent_uuid = None

        self.status_str = 'Fresh Server!'
        self.wait_time = 30.0 #seconds
        
        
        self.catch_protocol=None
        self.AL_status_str = ''
        self.data_manifest  = None
        self.AL_components = None
        self.components = None
        self.fixed_concs = None

        self.turbidity=None


        #XXX need to make deck inside this object because of 'different registries error in Pint
        self.reset_deck()
       
    def reset_deck(self):
        self.deck = AFL.automation.prepare.Deck()
        
    def add_container(self,name,slot):
        self.deck.add_container(name,slot)
        
    def add_catch(self,name,slot):
        self.deck.add_catch(name,slot)
        self.catch_loc=f"{slot}A1"
    
    def add_pipette(self,name,mount,tipracks):
        self.deck.add_pipette(name,mount,tipracks=tipracks)
        
    def send_deck_config(self,home=True):
        self.deck.init_remote_connection(
            self.prep_url.split(":")[0],
            home=home
        )
        self.deck.send_deck_config()
        
    def add_stock(self,stock_dict,loc):
        soln = AFL.automation.prepare.Solution.from_dict(stock_dict)
        self.deck.add_stock(soln,loc)
        
       
    def status(self):
        status = []
        status.append(f'Snapshots: {self.config["snapshot_directory"]}')
        status.append(f'Cameras: {self.camera_urls}')
        status.append(f'SAS: {self.sas_url}')
        status.append(f'Sample Wait Time: {self.wait_time}')
        status.append(f'AL_components: {self.AL_components}')
        status.append(f'{len(self.deck.stocks)} stocks loaded!')
        status.append(self.status_str)
        status.append(self.AL_status_str)
        status.append(f'Components: {self.components}')
        status.append(f'AL Components: {self.AL_components}')
        status.append(f'Fixed concs: {self.fixed_concs}')
        return status
    
    def update_status(self,value):
        self.status_str = value
        self.app.logger.info(value)

    def take_snapshot(self,prefix):
        now = datetime.datetime.now().strftime('%y%m%d-%H:%M:%S')
        for i,cam_url in enumerate(self.camera_urls):
            fname = self.config['snapshot_directory'] + '/' 
            fname += prefix
            fname += f'-{i}-'
            fname += now
            fname += '.jpg'

            try:
                r = requests.get(cam_url,stream=True)
                if r.status_code == 200:
                    with open(fname,'wb') as f:
                        r.raw.decode_content=True
                        shutil.copyfileobj(r.raw,f)
            except Exception as error:
                output_str  = f'take_snapshot failed with error: {error.__repr__()}\n\n'+traceback.format_exc()+'\n\n'
                self.app.logger.warning(output_str)

    def measure_SAS(self,sample):
        exposure = sample.get('exposure',None)
        sas_uuid = self.sas_client.enqueue(task_name='expose',name=sample['name'],block=True,exposure=exposure)
        self.sas_client.wait(sas_uuid)
        return sas_uuid

    def measure_spec(self,sample):
        if self.spec_client is None:
            raise ValueError('No spec client to do spec measurement with! Restart server with spec IP:port specified!')

        self.spec_client.set_config(
                filepath=self.config['spec_data_path'],
                filename=sample['name']
            )
        spec_uuid = self.spec_client.enqueue(task_name='collectContinuous',duration=sample['duration'],interactive=False)
        return spec_uuid

    def sample(self,sample):
        return self.process_sample(sample)
        
    def process_sample(self,sample):
        name = sample['name']

        targets = set()
        for task in sample['prep_protocol']:
            if 'target' in task['source'].lower():
                targets.add(task['source'])
            if 'target' in task['dest'].lower():
                targets.add(task['dest'])

        for task in sample['catch_protocol']:
            if 'target' in task['source'].lower():
                targets.add(task['source'])
            if 'target' in task['dest'].lower():
                targets.add(task['dest'])

        target_map = {}
        for t in targets:
            prep_target = self.prep_client.enqueue(task_name='get_prep_target',interactive=True)['return_val']
            target_map[t] = prep_target

        for i,task in enumerate(sample['prep_protocol']):
            #if the well isn't in the map, just use the well
            task['source'] = target_map.get(task['source'],task['source'])
            task['dest'] = target_map.get(task['dest'],task['dest'])
            if i==0:
                task['force_new_tip']=True
            if i==(len(sample['prep_protocol'])-1):#last prepare
                task['drop_tip']=False
            self.prep_uuid = self.prep_client.transfer(**task)
 
        if self.rinse_uuid is not None:
            self.update_status(f'Waiting for rinse...')
            self.load_client.wait(self.rinse_uuid)
            self.update_status(f'Rinse done!')

        #calibrate sensor to avoid drift
        self.load_client.enqueue(task_name='calibrate_sensor')

        self.turb_client.enqueue(task_name='measure',plotting=True,set_empty=True)

        self.update_status(f'Cell is clean, measuring empty cell scattering...')
        empty = {}
        empty['name']     = 'MT-'+sample['name']+'-spec.h5'
        empty['duration'] = 5
        self.spec_uuid = self.measure_spec(empty)
        self.spec_client.wait(self.spec_uuid)

        ## TEMPORARILY REMOVED TO SPEED UP MEASUREMENT!!!
        # empty = {}
        # empty['name'] = 'MT-'+sample['name']
        # empty['exposure'] = sample['empty_exposure']
        # empty['wiggle'] = False
        # self.sas_uuid = self.measure_SAS(empty)
        # self.sas_client.wait(self.sas_uuid)
            
        if self.prep_uuid is not None: 
            self.prep_client.wait(self.prep_uuid)
            self.take_snapshot(prefix = f'02-after-prep-{name}')

        
        self.update_status(f'Queueing sample {name} load into syringe loader')
        for task in sample['catch_protocol']:
            #if the well isn't in the map, just use the well
            task['source'] = target_map.get(task['source'],task['source'])
            task['dest'] = target_map.get(task['dest'],task['dest'])
            self.catch_uuid = self.prep_client.transfer(**task)
        
        if self.catch_uuid is not None:
            self.update_status(f'Waiting for sample prep/catch of {name} to finish: {self.catch_uuid}')
            self.prep_client.wait(self.catch_uuid)
            self.take_snapshot(prefix = f'03-after-catch-{name}')

        #homing robot to try to mitigate drift problems
        self.prep_client.home()

        self.update_status(f'Loading sample into spectrometer cell...')
        self.load_uuid = self.load_client.enqueue(task_name='loadSample',load_dest_label='afterSPEC')
        self.load_client.wait(self.load_uuid)
        self.take_snapshot(prefix = f'05-after-load-spec-{name}')

        self.update_status(f'Sample is loaded into cell, asking the spectrometer for exposure...')
        spec_sample = {}
        spec_sample['name'] = sample['name']+'-beforeSAS-spec.h5'
        spec_sample['duration'] = 5
        self.spec_uuid = self.measure_spec(spec_sample)
        self.spec_client.wait(self.spec_uuid)

        self.update_status(f'Loading sample into SAS cell...(while taking spec)')
        spec_sample = {}
        spec_sample['name'] = sample['name']+'-SPECToSAS-spec.h5'
        spec_sample['duration'] = 15
        self.load_uuid = self.load_client.enqueue(task_name='advanceSample',load_dest_label='afterSANS')
        self.spec_uuid = self.measure_spec(spec_sample)
        self.load_client.wait(self.load_uuid)
        self.take_snapshot(prefix = f'06-after-load-sans-{name}')

        #measure turbidity
        self.turbidity = self.turb_client.enqueue(task_name='measure',plotting=True,interactive=True)['return_val'][0]

        self.update_status(f'Sample is loaded into cell, asking the SAS instrument for exposure...')
        self.sas_uuid = self.measure_SAS(sample)
        self.sas_client.wait(self.sas_uuid)

        self.update_status(f'Cleaning up sample {name}...')
        self.rinse_uuid = self.load_client.enqueue(task_name='rinseCell')

        self.update_status(f'Waiting for instrument to measure scattering of {name} with UUID {self.sas_uuid}...')
        self.take_snapshot(prefix = f'07-after-measure-{name}')
            
        self.update_status(f'All done for {name}!')


    @Driver.unqueued()
    def stop_active_learning(self):
        self.stop_AL = True
        
    def set_catch_protocol(self,**kwargs): 
        self.catch_protocol = AFL.automation.prepare.PipetteAction(**kwargs)
    
    
    def mfrac_to_mass(self,mass_fractions,specified_conc,sample_volume,output_units='mg'):
        if not (len(mass_fractions)==3):
            raise ValueError('Only ternaries are currently supported. Need to pass three mass fractions')
            
        if len(specified_conc)>1:
            raise ValueError('Only one concentration should be fixed!')
        specified_component = list(specified_conc.keys())[0]
        
        components = list(mass_fractions.keys())
        components.remove(specified_component)
        
        xB = mass_fractions[components[0]]*units('')
        xC = mass_fractions[components[1]]*units('')
        XB = xB/(1-xB)
        XC = xC/(1-xC)
        
        mA = (specified_conc[specified_component]*sample_volume)
        mC = mA*(XC+XB*XC)/(1-XB*XC)
        mB = XB*(mA+mC)
        
        mass_dict = {}
        mass_dict[specified_component] = (mA).to(output_units)
        mass_dict[components[0]] = (mB).to(output_units)
        mass_dict[components[1]] = (mC).to(output_units)
        return mass_dict
    
    def active_learning_loop(self,**kwargs):
        # grid must have units for each component specified, hopefully this will progate to next_sample....
        self.components     = kwargs['components']
        self.AL_components  = kwargs['AL_components']
        self.AL_kwargs      = kwargs.get('AL_kwargs',{})
        self.fixed_concs    = kwargs.get('fixed_concs',{})
        sample_volume       = kwargs['sample_volume']
        sample_volume_units = kwargs['sample_volume_units']
        sample_volume       = sample_volume*units(sample_volume_units)
        exposure            = kwargs['exposure']
        empty_exposure      = kwargs['empty_exposure']
        predict             = kwargs.get('predict',True)
        start_new_manifest  = kwargs.get('start_new_manifest',False)
        ternary             = kwargs['ternary']#should be bool
        pre_run_list        = copy.deepcopy(kwargs.get('pre_run_list',[]))

        mix_order = kwargs.get('mix_order',None)
        custom_stock_settings = kwargs.get('custom_stock_settings',None)

        AL_manifest_path = pathlib.Path(self.config['AL_manifest_file'])
        data_path = pathlib.Path(self.config['csv_data_path'])
        
        self.stop_AL = False
        while not self.stop_AL:
            self.app.logger.info(f'Starting new AL loop')
            #####################
            ## GET NEXT SAMPLE ##
            #####################
            if pre_run_list:
                self.next_sample = None
                next_sample_dict = pre_run_list.pop(0)
            else:
                self.next_sample = self.agent_client.get_object('next_sample')
                next_sample_dict = self.next_sample.squeeze()
                if 'grid' in next_sample_dict:
                    next_sample_dict = next_sample_dict.reset_coords('grid',drop=True)
                next_sample_dict = next_sample_dict.to_pandas().to_dict()
                next_sample_dict = {k:{'value':v,'units':self.next_sample.attrs[k+'_units']} for k,v in next_sample_dict.items()}
            self.app.logger.info(f'Preparing to make next sample: {next_sample_dict}')

            ##############################
            ## PROTOCOL FOR NEXT SAMPLE ##
            ##############################
            if ternary:
                conc_spec = {}
                for name,value in self.fixed_concs.items():
                    conc_spec[name] = value['value']*units(value['units'])
                
                mass_dict = self.mfrac_to_mass(
                    mass_fractions=next_sample_dict,
                    specified_conc=conc_spec,
                    sample_volume=sample_volume,
                    output_units='mg')
            else:
                #assume concs for now...
                if len(next_sample_dict)<(len(self.components)-1):
                    raise ValueError('System under specified...')

                mass_dict = {}
                for name,comp in next_sample_dict.items():
                    mass_dict[name] = (comp['value']*units(comp['units'])*sample_volume).to('mg')
            
            self.target = AFL.automation.prepare.Solution('target',self.components)
            self.target.volume = sample_volume
            for k,v in mass_dict.items():
                self.target[k].mass = v
            self.target.volume = sample_volume


            ######################
            ## MAKE NEXT SAMPLE ##
            ######################
            self.deck.reset_targets()
            self.deck.add_target(self.target,name='target')
            self.deck.make_sample_series(reset_sample_series=True)
            self.deck.validate_sample_series(tolerance=0.15)
            self.deck.make_protocol(only_validated=False)
            if (mix_order is None) or (custom_stock_settings is None):
                warnings.warn('No mix_order or custom_stock_settings applied as mix_order={mix_order} and custom_stock_settings={custom_stock_settings}')
            else: 
                self.fix_protocol_order(mix_order,custom_stock_settings)
            self.sample,validated = self.deck.sample_series[0]
            self.app.logger.info(self.deck.validation_report)
            
            if validated:
                self.app.logger.info(f'Validation PASSED')
                self.AL_status_str = 'Last sample validation PASSED'
            else:
                self.app.logger.info(f'Validation FAILED')
                self.AL_status_str = 'Last sample validation FAILED'
            self.app.logger.info(f'Making next sample with mass fraction: {self.sample.target_check.mass_fraction}')
            
            self.catch_protocol.source = self.sample.target_loc
            
            sample_uuid = str(uuid.uuid4())
            sample_name = f'AL_{self.config["data_tag"]}_{sample_uuid[-8:]}'

            sample_data = self.set_sample(sample_name=sample_name,sample_uuid=sample_uuid)
            self.prep_client.enqueue(task_name='set_sample',**sample_data)
            self.load_client.enqueue(task_name='set_sample',**sample_data)
            self.sas_client.enqueue(task_name='set_sample',**sample_data)
            self.agent_client.enqueue(task_name='set_sample',**sample_data)
            self.turb_client.enqueue(task_name='set_sample',**sample_data)
            if self.spec_client is not None:
                self.spec_client.enqueue(task_name='set_sample',**sample_data)

            self.process_sample(
                    dict(
                        name=sample_name,
                        prep_protocol = self.sample.emit_protocol(),
                        catch_protocol =[self.catch_protocol.emit_protocol()],
                        volume = self.catch_protocol.volume/1000.0,
                        exposure = exposure,
                        empty_exposure = empty_exposure
                )
            )


            ################################
            ## VERIFY SAMPLE TRANSMISSION ##
            ################################
            h5_path = data_path / (sample_name+'.h5')
            with h5py.File(h5_path,'r') as h5:
                SAS_transmission = h5['entry/sample/transmission'][()]
  
            if SAS_transmission>self.config['max_sample_transmission']:
                self.update_status(f'Last sample missed! (Transmission={SAS_transmission})')
                self.app.logger.info('Dropping this sample from AL and hoping the next one hits...')
                continue
            else:
                self.update_status(f'Last Sample success! (Transmission={SAS_transmission})')

            ###############################
            ## GRAB SPEC DATA FROM TILED ##
            ###############################
            res = self.tiled_cat.search(Eq('sample_uuid',sample_uuid)).search(Eq('task_name','collectContinuous'))
            res = [v for k,v in res.items() if 'beforeSAS' in v.metadata['driver_config']['filename']]
            if not (len(res)==1):
                raise ValueError(f'Bad tiled return: Found {len(res)} results matching query')
            wl = res[0].metadata['wavelength']      
            data = res[0].read()
            spec_measurement = xr.DataArray(data,dims=['spec_time','wavelength'],coords={'wavelength':wl})

            ##################################
            ## BUILD DATASET FOR NEW SAMPLE ##
            ##################################
            data_fname = sample_name+'_chosen_r1d.csv'
            SAS_measurement = pd.read_csv(data_path/data_fname,sep=',',comment='#',header=None,names=['q','I'],usecols=[0,1]).set_index('q').squeeze().to_xarray()
            SAS_measurement = SAS_measurement.dropna('q')

            self.new_data                     = xr.Dataset()
            self.new_data['fname']            = data_fname
            self.new_data['SAS']              = SAS_measurement
            self.new_data['turbidity']        = -self.turbidity # now high turbidity is positive
            self.new_data['spec']             = spec_measurement.median('spec_time')
            self.new_data['spec_error']       = ('wavelength',iqr(spec_measurement,axis=0))
            self.new_data['label']            = -1
            self.new_data['validated']        = validated
            self.new_data['SAS_transmission'] = SAS_transmission
            self.new_data['sample_uuid']      = sample_uuid

            sample_composition = {}
            if ternary:
                total = 0
                for component in self.AL_components:
                    mf = self.sample.target_check.mass_fraction[component].magnitude
                    self.new_data[component] = mf
                    total+=mf
                for component in self.AL_components:
                    self.new_data[component] = self.new_data[component]/total

                    #for tiled
                    sample_composition['ternary_mfrac_'+component] = self.sample.target_check.concentration[component].to("mg/ml").magnitude
            else:
                for component in self.AL_components:
                    # if component=='P188':
                    #     self.new_data[component] = 10.0
                    #     self.new_data[component].attrs['units'] = 'mg/ml'
                    #     real_value = self.sample.target_check.concentration[component].to("mg/ml").magnitude
                    #     self.new_data[component].attrs['real_value'] = real_value
                    #     self.new_data[component].attrs['explanation'] = 'Fixing P188 value to 10.0 mg/ml for GP reasons'
                    #     print(f'Fixing P188={real_value} to 10.0')
                    # else:
                    #     real_value = -1
                    #     self.new_data[component] = self.sample.target_check.concentration[component].to("mg/ml").magnitude
                    #     self.new_data[component].attrs['units'] = 'mg/ml'
                    real_value = -1
                    self.new_data[component] = self.sample.target_check.concentration[component].to("mg/ml").magnitude
                    self.new_data[component].attrs['units'] = 'mg/ml'


                    #for tiled
                    sample_composition['conc_'+component] = self.sample.target_check.concentration[component].to("mg/ml").magnitude
                
            for component in self.components:
                self.new_data['mfrac_'+component] = self.sample.target_check.mass_fraction[component].magnitude
                self.new_data['mass_'+component]  = self.sample.target_check[component].mass.to('mg').magnitude
                self.new_data['mass_'+component].attrs['units'] = 'mg'

                #for tiled
                sample_composition['mfrac_'+component] = self.sample.target_check.mass_fraction[component].magnitude
                sample_composition['mass_'+component]  = self.sample.target_check[component].mass.to('mg').magnitude

            sample_composition['components'] = self.components
            sample_composition['conc_units'] = 'mg/ml'
            sample_composition['mass_units'] = 'mg'
            if self.data is not None:
                self.data['sample_composition'] = sample_composition
                self.data['time'] = datetime.datetime.now().strftime('%m/%d/%y %H:%M:%S-%f %Z%z')
                self.data.finalize()

            
            ########################
            ## UPDATE AL MANIFEST ##
            ########################
            if start_new_manifest and AL_manifest_path.exists():
                self.update_status(f'Backing up old and then starting new manifest...')
                AL_manifest_backup = xr.load_dataset(AL_manifest_path)
                i=0
                while True:
                    AL_manifest_backup_path = pathlib.Path(str(AL_manifest_path)+f'.bak{i}')
                    if AL_manifest_backup_path.exists():
                        i+=1
                    else:
                        print(f'--> Writing backup AL file to {AL_manifest_backup_path}')
                        AL_manifest_backup.to_netcdf(AL_manifest_backup_path)
                        break

                self.AL_manifest = self.new_data.expand_dims('sample')
                start_new_manifest=False

            elif AL_manifest_path.exists():
                self.update_status(f'Updating manifest...')
                self.AL_manifest = xr.load_dataset(AL_manifest_path)

                if 'grid' in self.AL_manifest.dims:
                    self.AL_manifest = self.AL_manifest.drop_dims('grid')

                self.AL_manifest = xr.concat([self.AL_manifest,self.new_data],dim='sample')

            else:
                self.update_status(f'Starting new manifest...')
                self.AL_manifest = self.new_data.expand_dims('sample')
            
            self.AL_manifest.attrs.update(self.AL_kwargs)
            self.AL_manifest.attrs['AL_data'] = 'SAS'
            self.AL_manifest.attrs['components'] = self.AL_components

            #self.AL_manifest  = self.AL_manifest.afl.comp.add_grid(pts_per_row=pts_per_row)

            self.AL_manifest.to_netcdf(AL_manifest_path)

            ################
            ## TRIGGER AL ##
            ################
            self.update_status(f'Triggering agent server...')
            if len(pre_run_list)==0:
                if predict:
                    self.agent_uuid = self.agent_client.enqueue(task_name='predict',datatype='nc')
                
                    # wait for AL
                    self.app.logger.info(f'Waiting for agent...')
                    self.agent_client.wait(self.agent_uuid)
                else:#used for intialization
                    return
            
            
    def fix_protocol_order(self,mix_order,custom_stock_settings):
        mix_order = [self.deck.get_stock(i) for i in mix_order]
        mix_order_map = {loc:new_index for new_index,(stock,loc) in enumerate(mix_order)}
        for sample,validated in self.deck.sample_series:
            # if not validated:
            #     continue
            old_protocol = sample.protocol
            ordered_indices = list(map(lambda x: mix_order_map.get(x.source),sample.protocol))
            argsort = np.argsort(ordered_indices)
            new_protocol = list(map(sample.protocol.__getitem__,argsort))
            time_patched_protocol = []
            for entry in new_protocol:
                if entry.source in custom_stock_settings:
                    for setting,value in custom_stock_settings[entry.source].items():
                        entry.__setattr__(setting,value)
                time_patched_protocol.append(entry)
            sample.protocol = time_patched_protocol
      
   






   



"""
A collection of helper methods/classes
"""

import inspect
from typing import Any, Dict

from AFL.double_agent.PipelineOp import PipelineOp


def listify(obj):
    """Make any input an iterable list

    The primary use case is to handle inputs that are sometimes length=1 and not always passed as lists. In particular,
    this method handles string inputs which we do not want to iterate over.

    Example
    -------
    ```python
    def my_func(input):
        for i in listify(input):
            print(i)

    In[1]: my_func(1)
    Out[2]:
    1

    In[1]: my_func([1,2])
    Out[2]:
    1
    2

    In[1]: my_func('test')
    Out[2]:
    'test'
    ```

    In the last example, without listify the result would have been t,e,s,t on newlines.
    """
    if isinstance(obj, str) or not hasattr(obj, "__iter__"):
        obj = [obj]
    return obj


def extract_parameters(op: PipelineOp, method: str = "__init__") -> Dict:
    """Attempt to reconstruct the input parameters for a object's constructor

    Parameters
    ----------
    op: Any
        Technically any Python object but targeted at PipelineOps

    method: str
        While method to try to reconstruct. Typically, __init__
    """
    # grab base signature and default parameters
    signature = inspect.signature(getattr(op, method))

    parameters = {k: v.default for k, v in signature.parameters.items()}
    for k, v in signature.parameters.items():
        if k in ('input_variables','output_variables') and k not in op.__dict__:
            parameters[k] = op.__dict__.get(k[:-1], v)
        else:
            parameters[k] = op.__dict__.get(k, v)

    return parameters



"""
Boundary analysis tools for phase diagram and compositional spaces.

This module provides tools for analyzing and comparing boundaries in phase diagrams
and compositional spaces. It includes functionality for creating concave hulls around
phase regions and computing similarity scores between different boundary sets.

Key features:
- Concave hull generation for phase regions
- Boundary comparison metrics
- Support for 2D compositional spaces
- Visualization tools for boundary analysis
- Handling of multi-phase systems
"""

from typing import Dict, List, Optional, Generator
from typing_extensions import Self

import numpy as np
import xarray as xr

from sklearn.metrics import pairwise_distances_argmin_min  # type: ignore

from shapely import MultiPoint, Point, concave_hull  # type: ignore
from shapely.geometry import mapping, shape  # type: ignore
from shapely.errors import GEOSException

from plotly import graph_objects as go

from ast import literal_eval

from AFL.double_agent.PipelineOp import PipelineOp


# Shapely types to skip over for boundary score
EXCLUDED_HULL_TYPES = ["Point", "LineString"]


class ConcaveHull(PipelineOp):
    """Creates concave hulls around points in 2D space.
    
    This class generates concave hulls (alpha shapes) around sets of points,
    useful for identifying phase boundaries in compositional spaces. It can
    handle multiple phases and provides options for hull generation parameters.

    Parameters
    ----------
    input_variable : str
        The name of the variable containing the points to create hulls around

    output_prefix : str
        Prefix to use for output variable names

    hull_tracing_ratio : float, default=0.2
        Controls the concavity of the hull. Lower values create tighter hulls

    drop_phases : Optional[List], default=None
        List of phase labels to exclude from hull generation

    component_dim : str, default="components"
        Name of the dimension containing component coordinates

    label_variable : Optional[str], default=None
        Name of variable containing phase labels. If None, all points are treated
        as one phase

    segmentize_length : float, default=0.025
        Length to use when segmenting hull boundaries for smoother visualization

    allow_holes : bool, default=False
        Whether to allow holes in the generated hulls

    name : str, default="ConcaveHull"
        Name to use when added to a Pipeline
    """

    def __init__(
        self,
        input_variable: str,
        output_prefix: str,
        hull_tracing_ratio: float = 0.2,
        drop_phases: Optional[List] = None,
        component_dim: str = "components",
        label_variable: Optional[str] = None,
        segmentize_length: float = 0.025,
        allow_holes: bool = False,
        name: str = "ConcaveHull",
    ) -> None:
        super().__init__(
            name=name,
            input_variable=input_variable,
            output_variable=output_prefix + "_" + "hulls",
            output_prefix=output_prefix,
        )

        self.hull_tracing_ratio = hull_tracing_ratio
        self.component_dim = component_dim
        self.label_variable = label_variable
        self.segmentize_length = segmentize_length
        self.label_variable = label_variable
        self.drop_phases = drop_phases
        self.allow_holes = allow_holes

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Generate concave hulls for each phase in the dataset.
        
        Creates concave hulls around points grouped by their phase labels.
        Handles 2D data only and includes options for hull parameters.

        Parameters
        ----------
        dataset : xr.Dataset
            The input dataset containing points and optional phase labels

        Returns
        -------
        Self
            The ConcaveHull instance with generated hulls

        Raises
        ------
        ValueError
            If the input data is not 2D
        """
        input_data = self._get_variable(dataset).transpose(..., self.component_dim)
        if input_data.sizes[self.component_dim] != 2:
            raise ValueError(
                """ConcaveHull only works for 2D data. If working with ternary, do a Barycentric transform first."""
            )

        if self.label_variable is None:
            dim_name = input_data.dims[0]
            labels = xr.DataArray(np.ones(input_data.sizes[dim_name]), dims=dim_name)
        else:
            labels = dataset[self.label_variable]

        if self.drop_phases is None:
            self.drop_phases = []

        hulls = {}
        for label, xy in input_data.groupby(labels):
            if label in self.drop_phases:
                continue
            self.xys = MultiPoint(xy.values)
            try:
                hull = concave_hull(
                    self.xys, ratio=self.hull_tracing_ratio, allow_holes=self.allow_holes
                )
            except GEOSException:
                hulls[label] = Point(
                    [-3.14, 3.14]
                )  # dummy point at pi-pi, "Point" type is in EXCLUDED_HULL_TYPES
            else:
                hull = hull.segmentize(self.segmentize_length)
                hulls[label] = hull

        # This is a little...gross, but I'm not sure of a better way to store the variable length arrays
        # associated with concave hulls. We could store each hull as its own DataArray but we would need
        # separate dimensions for each hull...which seems like a mess. The compromise is to store a string
        # representation of the hull
        hulls_serialized = {}
        for k, v in hulls.items():
            hulls_serialized[k] = str(mapping(v))
        hulls_xr = xr.Dataset(hulls_serialized).to_array(
            self._prefix_output("hull_labels")
        )

        self.output[self._prefix_output("hulls")] = hulls_xr
        self.output[self._prefix_output("hulls")].attrs[
            "description"
        ] = "String representation of shapely concave hulls"

        return self


def _get_xy(hull):
    """Extract xy coordinates from shapely object.
    
    Parameters
    ----------
    hull : shapely.geometry
        The shapely geometry object to extract coordinates from

    Returns
    -------
    tuple[list, list]
        The x and y coordinates of the hull boundary
    """
    try:
        x, y = hull.boundary.xy
    except NotImplementedError:
        if hasattr(hull.boundary, "geoms"):
            x = []
            y = []
            for geom in list(hull.boundary.geoms):
                xx, yy = geom.xy
                x.extend(xx)
                y.extend(yy)
        else:
            x, y = None, None
    return x, y


def _get_hulls(
    dataset: xr.Dataset, hull_variable: str
):  # -> Generator[object, (xr.Dataset,str), None]: # to-fix when I have wifi...
    """Iterate over a data variable of shapely hull strings and convert to shapely objects.

    Parameters
    ----------
    dataset: xr.Dataset
        Input dataset containing hull data

    hull_variable: str
        Name of variable containing a data variable of stringified shapely objects. Should have only
        one dimension with a coordinate of hull_labels.

    Yields
    ------
    tuple[str, shapely.geometry]
        Tuple of (hull_label, hull_object) for each hull in the dataset
    """
    dims = dataset[hull_variable].dims
    if len(dims) > 1:
        raise ValueError(
            f"The hulls variable {hull_variable} should only have one dimension. It has {dims}"
        )
    dim = dims[0]

    for hull_label in dataset[hull_variable][dim].values:
        hull_str = dataset[hull_variable].sel(**{dim: hull_label}).item()
        hull = shape(literal_eval(hull_str))
        yield hull_label, hull


def _calculate_perimeter_score(hull1, hull2):
    """Calculate a similarity score between two hull perimeters.
    
    Computes a score based on the average minimum distance between points
    on the perimeters of two hulls.

    Parameters
    ----------
    hull1 : shapely.geometry
        First hull to compare
    hull2 : shapely.geometry
        Second hull to compare

    Returns
    -------
    dict
        Dictionary containing score statistics and visualization data
    """
    hull1_xy = np.vstack(_get_xy(hull1)).T
    hull2_xy = np.vstack(_get_xy(hull2)).T

    idx1, dist1 = pairwise_distances_argmin_min(hull2_xy, hull1_xy, metric="euclidean")
    idx2, dist2 = pairwise_distances_argmin_min(hull1_xy, hull2_xy, metric="euclidean")

    # build index list of closest pairs
    idx1 = np.vstack([np.arange(idx1.shape[0]), idx1]).T
    idx2 = np.vstack([idx2, np.arange(idx2.shape[0])]).T

    dist = np.hstack([dist1, dist2])
    idx = np.vstack([idx1, idx2])

    # remove duplicate pairs from idx and dist
    unique = np.unique(idx, axis=0, return_index=True)[1]
    idx = idx[unique]
    dist = dist[unique]

    # build list of distance pair coordinates for plotting
    coord = []
    for i, j in idx:
        coord.extend([hull1_xy[j], hull2_xy[i], [np.nan, np.nan]])
    coord = np.array(coord)

    out = {
        "mean": dist.mean(),
        "std": dist.std(),
        "pair_dist": dist,
        "pair_idx": idx,
        "pair_coord": coord,
        "hull1_xy": hull1_xy,
        "hull2_xy": hull2_xy,
    }

    return out


def _make_boundary_plots(score_dict):
    """Create plotly visualization of hull boundaries and their comparison.
    
    Parameters
    ----------
    score_dict : dict
        Dictionary containing hull coordinates and comparison data

    Returns
    -------
    dict
        Dictionary of plotly traces for visualization
    """
    gt_xy = score_dict["hull1_xy"]
    xy = score_dict["hull2_xy"]
    pair_xy = score_dict["pair_coord"]

    out = {}
    out["GT"] = go.Scatter(
        x=gt_xy.T[0],
        y=gt_xy.T[1],
        mode="lines+markers",
        line={"color": "black"},
        marker={"symbol": "circle-open"},
        name="GT",
    )
    out["AL"] = go.Scatter(
        x=xy.T[0],
        y=xy.T[1],
        mode="lines+markers",
        line={"color": "green"},
        marker={"symbol": "triangle-up-open"},
        name="AL",
    )
    out["pairs"] = go.Scatter(
        x=pair_xy[:, 0], y=pair_xy[:, 1], line={"color": "red"}, opacity=0.5, name=None
    )
    return out


class BoundaryScore(PipelineOp):
    """Computes similarity scores between two sets of phase boundaries.
    
    This class calculates metrics comparing the boundaries of phase regions
    between two different datasets (e.g., ground truth vs. predicted phases).
    It provides both numerical scores and optional visualization.

    Parameters
    ----------
    gt_hull_variable : str
        Name of variable containing ground truth hull data

    al_hull_variable : str
        Name of variable containing comparison hull data

    output_prefix : str, default="boundary_score"
        Prefix to use for output variable names

    name : str, default="BoundaryScore"
        Name to use when added to a Pipeline

    plot : bool, default=False
        Whether to create visualizations of the boundary comparisons
    """

    def __init__(
        self,
        gt_hull_variable: str,
        al_hull_variable: str,
        output_prefix: str = "boundary_score",
        name: str = "BoundaryScore",
        plot: bool = False,
    ) -> None:
        super().__init__(
            name=name,
            input_variable=[gt_hull_variable, al_hull_variable],
            output_variable=[output_prefix + "_" + i for i in ["mean", "std"]],
            output_prefix=output_prefix,
        )

        self.gt_hull_variable = gt_hull_variable
        self.al_hull_variable = al_hull_variable
        self.plot = plot

    def calculate(self, dataset: xr.Dataset) -> Self:
        """Calculate boundary similarity scores between hull sets.
        
        Computes mean and standard deviation of distances between corresponding
        points on hull boundaries. Optionally creates visualizations of the
        comparison.

        Parameters
        ----------
        dataset : xr.Dataset
            Dataset containing both sets of hulls to compare

        Returns
        -------
        Self
            The BoundaryScore instance with computed scores
        """
        all_scores = []
        for gt_name, gt_hull in _get_hulls(dataset, self.gt_hull_variable):
            if gt_hull.geom_type in EXCLUDED_HULL_TYPES:
                continue
            for al_name, al_hull in _get_hulls(dataset, self.al_hull_variable):
                if al_hull.geom_type in EXCLUDED_HULL_TYPES:
                    continue

                try:
                    score = _calculate_perimeter_score(gt_hull, al_hull)
                except NotImplementedError:
                    continue

                score["GT"] = gt_name
                score["AL"] = al_name
                all_scores.append(score)

        # find best matches for each ground truth phase
        all_scores = sorted(all_scores, key=lambda x: x["mean"])
        best_scores: Dict = {}
        for score in all_scores:
            check1 = [
                (score["AL"] == value["AL"]) for key, value in best_scores.items()
            ]
            check2 = [
                (score["GT"] == value["GT"]) for key, value in best_scores.items()
            ]
            if not (any(check1) or any(check2)):
                best_scores[score["GT"]] = score

        if self.plot:
            for label, score_dict in best_scores.items():
                boundary_plots = _make_boundary_plots(score_dict)
                fig = go.FigureWidget(
                    layout={
                        "width": 500,
                        "height": 500,
                        "margin": dict(l=20, r=20, t=20, b=20),
                        "legend": dict(
                            yanchor="top", y=0.99, xanchor="left", x=0.01, bgcolor=None
                        ),
                    }
                )
                fig.add_trace(boundary_plots["GT"])
                fig.add_trace(boundary_plots["AL"])
                fig.add_trace(boundary_plots["pairs"])
                fig.show()

        self.output[self._prefix_output("mean")] = xr.Dataset(
            {k: v["mean"] for k, v in best_scores.items()}
        ).to_array("hull_label")
        self.output[self._prefix_output("mean")].attrs[
            "description"
        ] = "Mean of boundary score for all matched phases"

        self.output[self._prefix_output("std")] = xr.Dataset(
            {k: v["std"] for k, v in best_scores.items()}
        ).to_array("hull_label")
        self.output[self._prefix_output("std")].attrs[
            "description"
        ] = "Standard deviation of boundary score for all matched phases"

        return self



import copy
import pathlib
import uuid
from collections import defaultdict
from typing import Dict, Any, Optional, List, Union

import h5py  # type: ignore
import numpy as np
import pandas as pd
import xarray as xr

from AFL.automation.APIServer.Driver import Driver  # type: ignore
from AFL.double_agent.AutoSAS import SASModel, SASFitter


class AutoSAS_Driver(Driver):
    defaults = {}
    defaults["savepath"] = "/home/afl642/2402_DT_ISIS_path"
    defaults["q_min"] = 1e-2
    defaults["q_max"] = 1e-1
    defaults["resolution"] = None
    defaults["model_inputs"] = SASFitter.DEFAULT_MODEL_INPUTS
    defaults["fit_method"] = SASFitter.DEFAULT_FIT_METHOD

    def __init__(self):
        Driver.__init__(self, name="SAS_model_fitter", defaults=self.gather_defaults())
        self.status_str = "Fresh Server!"
        self.dropbox = dict()
        self.fitter = None
        print("self.data exists == :", self.data)

    def status(self):
        status = []
        status.append(self.status_str)
        return status

    def update_status(self, status):
        self.status_str = status
        if self.app is not None:
            self.app.logger(status)

    def set_sasdata(
        self,
        db_uuid: str,
        sample_dim: str = "sample",
        q_variable: str = "q",
        sas_variable: str = "I",
        sas_err_variable: str = "dI",
        sas_resolution_variable: Optional[str] = None,
    ) -> None:
        """
        Set the sasdata to be fit from a dropbox UUID
        
        Parameters
        ----------
        db_uuid: str
            Dropbox UUID to retrieve `xarray.Dataset` from
            
        sample_dim: str
            The `xarray` dimension containing each sample

        q_variable: str
            The name of the `xarray.Dataset` variable corresponding to the q values

        sas_variable: str
            The name of the `xarray.Dataset` variable corresponding to the scattering intensity to be fit

        sas_err_variable: str
            The name of the `xarray.Dataset` variable corresponding to the uncertainty in the scattering intensity

        sas_resolution_variable: Optional[str]
            The name of the `xarray.Dataset` variable corresponding to the resolution function
        """
        dataset = self.retrieve_obj(db_uuid)
        
        # Initialize the fitter
        self.fitter = SASFitter(
            model_inputs=self.config["model_inputs"],
            fit_method=self.config["fit_method"],
            q_min=self.config["q_min"],
            q_max=self.config["q_max"],
            resolution=self.config["resolution"]
        )
        
        # Set the SAS data in the fitter
        self.fitter.set_sasdata(
            dataset=dataset,
            sample_dim=sample_dim,
            q_variable=q_variable,
            sas_variable=sas_variable,
            sas_err_variable=sas_err_variable,
            sas_resolution_variable=sas_resolution_variable
        )

    def fit_models(self, parallel=False, fit_method=None):
        """
        Execute SAS model fitting using the fitter
        
        Parameters
        ----------
        parallel: bool
            NOT IMPLEMENTED! Flag for parallel processing
            
        fit_method: Optional[Dict]
            Custom fit method parameters
            
        Returns
        -------
        str:
            UUID of the fit calculation
        """
        if self.fitter is None:
            raise ValueError("No SAS data set. Use set_sasdata first.")
        
        # Update fit method if provided
        if fit_method is not None:
            self.config["fit_method"] = fit_method
            self.fitter.fit_method = fit_method
        
        # Perform the fitting
        fit_uuid, output_dataset = self.fitter.fit_models(parallel=parallel)
        
        # Store in data and deposit to dropbox
        for name, array in output_dataset.data_vars.items():
            self.data.add_array(name, array.values)
            
            if array.dims:
                for dim in array.dims:
                    if dim in output_dataset.coords:
                        dim_name = f"{name}_dim_{dim}"
                        self.data.add_array(dim_name, output_dataset[dim].values)
        
        # Store best fit info
        self.data.add_array("best_chisq", self.fitter.report["best_fits"]["lowest_chisq"])
        self.data.add_array("model_names", self.fitter.report["best_fits"]["model_name"])
        self.data.add_array("all_chisq", self.fitter.report["all_chisq"])
        self.data.add_array("probabilities", self.fitter.report["probabilities"])
        
        # Deposit the output dataset to the dropbox
        self.deposit_obj(obj=output_dataset, uid=fit_uuid)
        
        return fit_uuid

    def _writedata(self, data):
        filename = pathlib.Path(self.config["filename"])
        filepath = pathlib.Path(self.config["filepath"])
        print(f"writing data to {filepath/filename}")
        with h5py.File(filepath / filename, "w") as f:
            f.create_dataset(str(uuid.uuid1()), data=data)

_DEFAULT_PORT = 5058
if __name__ == '__main__':
    from AFL.automation.shared.launcher import *



from AFL.automation.APIServer.Client import Client
from AFL.automation.prepare.OT2Client import OT2Client
from AFL.automation.shared.utilities import listify
from AFL.automation.APIServer.Driver import Driver
from AFL.agent.AgentClient import AgentClient
from AFL.automation.shared.units import units
from AFL.automation.prepare.Sample import Sample

import warnings

import pandas as pd
import numpy as np 
import xarray as xr 
from math import ceil,sqrt
import json
import time
import requests
import shutil
import datetime
import traceback
import pathlib
import uuid
import copy
import os

import AFL.automation.prepare
import shutil
import h5py

from tiled.client import from_uri
from tiled.queries import Eq,Contains


class vSAS_NoDeck_AL_SampleDriver(Driver):
    defaults={}
    defaults['snapshot_directory'] = '/home/nistoroboto'
    defaults['csv_data_path'] = '/nfs/aux/chess/reduced_data/cycles/2022-1/id3b/beaucage-2324-D/analysis/'
    defaults['AL_manifest_file'] = '/nfs/aux/chess/reduced_data/cycles/2022-1/id3b/beaucage-2324-D/analysis/data_manifest.csv'
    defaults['data_tag'] = 'default'
    defaults['max_sample_transmission'] = 0.6
    def __init__(self,
            load_url,
            prep_url,
            sas_url,
            agent_url,
            tiled_url,
            spec_url=None,
            camera_urls = None,
            snapshot_directory =None,
            overrides=None, 
            ):

        Driver.__init__(self,name='vSAS_NoDeck_AL_SampleDriver',defaults=self.gather_defaults(),overrides=overrides)

        if not (len(load_url.split(':'))==2):
            raise ArgumentError('Need to specify both ip and port on load_url')

        if not (len(prep_url.split(':'))==2):
            raise ArgumentError('Need to specify both ip and port on prep_url')

        if not (len(agent_url.split(':'))==2):
            raise ArgumentError('Need to specify both ip and port on agent_url')

        self.app = None
        self.name = 'vSAS_AL_SampleDriver'

        #prepare samples
        # self.prep_url = prep_url
        # self.prep_client = OT2Client(prep_url.split(':')[0],port=prep_url.split(':')[1])
        # self.prep_client.login('SampleServer_PrepClient')
        # self.prep_client.debug(False)

        #load samples
        # self.load_client = Client(load_url.split(':')[0],port=load_url.split(':')[1])
        # self.load_client.login('SampleServer_LoadClient')
        # self.load_client.debug(False)
 
        #measure samples
        self.sas_url = sas_url
        self.sas_client = Client(sas_url.split(':')[0],port=sas_url.split(':')[1])
        self.sas_client.login('SampleServer_SASClient')
        self.sas_client.debug(False)
        
        # #vSAS additional stuff
        # netcdf_path = '/Users/drs18/Documents/multimodal-dev/phasemap_P188_2D_MultiModal_UCB_noThomp_FixedP188_30524b88-00f7-4606-9c4d-57ad7880f95e.nc'
        # model_ds = xr.load_dataset(netcdf_path)
        # self.sas_client.set_driver_object(dataset=model_ds)
        # self.sas_client.set_driver_object(clustered=True)
        # self.sas_client.enqueue(task_name='load_model_dataset')
        # self.sas_client.enqueue(task_name='generate_model')
        # self.sas_client.enqueue(task_name='train_model',niter=1001)
        
        
        #agent
        self.agent_url = agent_url
        self.agent_client = AgentClient(agent_url.split(':')[0],port=agent_url.split(':')[1])
        self.agent_client.login('SampleServer_AgentClient')
        self.agent_client.debug(False)

        if spec_url is not None:
            self.spec_url = spec_url
            self.spec_client = Client(spec_url.split(':')[0],port=spec_url.split(':')[1])
            self.spec_client.login('SampleServer_LSClient')
            self.spec_client.debug(False)
            self.spec_client.enqueue(task_name='setExposure',time=0.1)
        else:
            self.spec_client = None

        # start tiled catalog connection
        self.tiled_cat = from_uri(tiled_url,api_key=os.environ['TILED_API_KEY'])
    
        if camera_urls is None:
            self.camera_urls = []
        else:
            self.camera_urls = camera_urls

        if snapshot_directory is not None:
            self.config['snapshot_directory'] = snapshot_directory

        self.rinse_uuid = None
        self.prep_uuid = None
        self.catch_uuid = None
        self.agent_uuid = None

        self.status_str = 'Fresh Server!'
        self.wait_time = 30.0 #seconds
        
        
        self.catch_protocol=None
        self.AL_status_str = ''
        self.data_manifest  = None
        self.AL_components = None
        self.components = None


       
    def status(self):
        status = []
        status.append(f'Snapshots: {self.config["snapshot_directory"]}')
        status.append(f'Cameras: {self.camera_urls}')
        status.append(f'SAS: {self.sas_url}')
        status.append(f'Sample Wait Time: {self.wait_time}')
        status.append(f'AL_components: {self.AL_components}')
        status.append(self.status_str)
        status.append(self.AL_status_str)
        status.append(f'Components: {self.components}')
        status.append(f'AL Components: {self.AL_components}')
        return status
    
    def update_status(self,value):
        self.status_str = value
        self.app.logger.info(value)

    def measure(self,sample):
        exposure = sample.get('exposure',None)

        if self.spec_client is not None:
            self.spec_client.set_config(
                    filepath='/home/pi/2305_SINQ_data_reduced/',
                    filename=f'{sample["name"]}-beforeSAS-spec.h5'
                )
            self.spec_client.enqueue(task_name='collectContinuous',duration=5,interactive=False)

        ### the point in which the virtual SAS data driver will produce a spectrum and store it to tiled
        sas_uuid = self.sas_client.enqueue(task_name='expose',name=sample['name'],block=True,exposure=exposure)
        self.sas_client.wait(sas_uuid)

        if self.spec_client is not None:
            self.spec_client.set_config(
                    filepath='/home/pi/2305_SINQ_data_reduced/',
                    filename=f'{sample["name"]}-afterSAS-spec.h5'
                )
            spec_uuid = self.spec_client.enqueue(task_name='collectContinuous',duration=5,interactive=False)
            self.spec_client.wait(spec_uuid)

        return sas_uuid

    def sample(self,sample):
        return self.process_sample(sample)
        
    def process_sample(self,sample):
        name = sample['name']

        targets = set()
        for task in sample['prep_protocol']:
            if 'target' in task['source'].lower():
                targets.add(task['source'])
            if 'target' in task['dest'].lower():
                targets.add(task['dest'])

        for task in sample['catch_protocol']:
            if 'target' in task['source'].lower():
                targets.add(task['source'])
            if 'target' in task['dest'].lower():
                targets.add(task['dest'])

        target_map = {}
        for t in targets:
            prep_target = self.prep_client.enqueue(task_name='get_prep_target',interactive=True)['return_val']
            target_map[t] = prep_target

        for i,task in enumerate(sample['prep_protocol']):
            #if the well isn't in the map, just use the well
            task['source'] = target_map.get(task['source'],task['source'])
            task['dest'] = target_map.get(task['dest'],task['dest'])
            if i==0:
                task['force_new_tip']=True
            if i==(len(sample['prep_protocol'])-1):#last prepare
                task['drop_tip']=False
            self.prep_uuid = self.prep_client.transfer(**task)
 
        if self.rinse_uuid is not None:
            self.update_status(f'Waiting for rinse...')
            self.load_client.wait(self.rinse_uuid)
            self.update_status(f'Rinse done!')

        if self.spec_client is not None:
            self.spec_client.set_config(
                    filepath='/home/pi/2305_SINQ_data_reduced/',
                    filename=f'{sample["name"]}-MT-spec.h5'
                )
            self.spec_client.enqueue(task_name='collectContinuous',duration=5,interactive=False)

        if sample['empty_exposure']>0.0:
            self.update_status(f'Cell is clean, measuring empty cell scattering...')
            empty = {}
            empty['name'] = 'MT-'+sample['name']
            empty['exposure'] = sample['empty_exposure']
            empty['wiggle'] = False
            self.sas_uuid = self.measure(empty)
            self.sas_client.wait(self.sas_uuid)
                
        if self.prep_uuid is not None: 
            self.prep_client.wait(self.prep_uuid)
            self.take_snapshot(prefix = f'02-after-prep-{name}')

        
        self.update_status(f'Queueing sample {name} load into syringe loader')
        for task in sample['catch_protocol']:
            #if the well isn't in the map, just use the well
            task['source'] = target_map.get(task['source'],task['source'])
            task['dest'] = target_map.get(task['dest'],task['dest'])
            self.catch_uuid = self.prep_client.transfer(**task)
        
        if self.catch_uuid is not None:
            self.update_status(f'Waiting for sample prep/catch of {name} to finish: {self.catch_uuid}')
            self.prep_client.wait(self.catch_uuid)
            self.take_snapshot(prefix = f'03-after-catch-{name}')

        #homing robot to try to mitigate drift problems
        self.prep_client.home()

        if self.spec_client is not None:
            self.spec_client.set_config(
                    filepath='/home/pi/2305_SINQ_data_reduced/',
                    filename=f'{sample["name"]}-duringLoad-spec.h5'
                )
            self.spec_client.enqueue(task_name='collectContinuous',duration=15,interactive=False)
        
        self.load_uuid = self.load_client.enqueue(task_name='loadSample',sampleVolume=sample['volume'])
        self.update_status(f'Loading sample into cell: {self.load_uuid}')
        self.load_client.wait(self.load_uuid)
        self.take_snapshot(prefix = f'05-after-load-{name}')
        
        self.update_status(f'Sample is loaded, asking the instrument for exposure...')
        self.sas_uuid = self.measure(sample)

        self.update_status(f'Cleaning up sample {name}...')
        self.rinse_uuid = self.load_client.enqueue(task_name='rinseCell')

        self.update_status(f'Waiting for instrument to measure scattering of {name} with UUID {self.sas_uuid}...')
        self.sas_client.wait(self.sas_uuid)
        self.take_snapshot(prefix = f'06-after-measure-{name}')
            
        self.update_status(f'All done for {name}!')
        
    def mfrac_to_mass(self,mass_fractions,specified_conc,sample_volume,output_units='mg'):
        if not (len(mass_fractions)==3):
            raise ValueError('Only ternaries are currently supported. Need to pass three mass fractions')
            
        if len(specified_conc)>1:
            raise ValueError('Only one concentration should be fixed!')
        specified_component = list(specified_conc.keys())[0]
        
        components = list(mass_fractions.keys())
        components.remove(specified_component)
        
        xB = mass_fractions[components[0]]*units('')
        xC = mass_fractions[components[1]]*units('')
        XB = xB/(1-xB)
        XC = xC/(1-xC)
        
        mA = (specified_conc[specified_component]*sample_volume)
        mC = mA*(XC+XB*XC)/(1-XB*XC)
        mB = XB*(mA+mC)
        
        mass_dict = {}
        mass_dict[specified_component] = (mA).to(output_units)
        mass_dict[components[0]] = (mB).to(output_units)
        mass_dict[components[1]] = (mC).to(output_units)
        return mass_dict


    @Driver.unqueued()
    def stop_active_learning(self):
        self.stop_AL = True
    
    def take_snapshot(self,*args,**kwargs):
        pass
    
    def active_learning_loop(self,**kwargs):
        # grid must have units for each component specified, hopefully this will progate to next_sample....
        self.components     = kwargs['components']
        self.AL_components  = kwargs['AL_components']
        self.AL_kwargs      = kwargs.get('AL_kwargs',{})
        self.fixed_concs    = kwargs.get('fixed_concs',{})
        predict             = kwargs.get('predict',True)
        start_new_manifest  = kwargs.get('start_new_manifest',False)
        ternary             = kwargs['ternary']#should be bool
        pre_run_list        = copy.deepcopy(kwargs.get('pre_run_list',[]))

        AL_manifest_path = pathlib.Path(self.config['AL_manifest_file'])
        data_path = pathlib.Path(self.config['csv_data_path'])
        
        self.stop_AL = False
        while not self.stop_AL:
            self.app.logger.info(f'Starting new AL loop')
            #####################
            ## GET NEXT SAMPLE ##
            #####################
            if pre_run_list:
                self.next_sample = None
                next_sample_dict = pre_run_list.pop(0)
            else:
                self.next_sample = self.agent_client.get_object('next_sample')
                next_sample_dict = self.next_sample.squeeze()
                if 'grid' in next_sample_dict:
                    next_sample_dict = next_sample_dict.reset_coords('grid',drop=True)
                next_sample_dict = next_sample_dict.to_pandas().to_dict()
                next_sample_dict = {k:{'value':v,'units':self.next_sample.attrs[k+'_units']} for k,v in next_sample_dict.items()}
            self.app.logger.info(f'Preparing to make next sample: {next_sample_dict}')
                    
            
            ##############################
            ## PROTOCOL FOR NEXT SAMPLE ##
            ##############################
            if ternary and (len(self.components)==4):
                conc_spec = {}
                for name,value in self.fixed_concs.items():
                    conc_spec[name] = value['value']*units(value['units'])
                
                mass_dict = self.mfrac_to_mass(
                    mass_fractions=next_sample_dict,
                    specified_conc=conc_spec,
                    sample_volume=1*units("ml"),
                    output_units='mg')
            elif ternary and (len(self.components)==3):
                mass_dict = {}
                sample_mass = 1*units('mg')
                for name,comp in next_sample_dict.items():
                    mass_dict[name] = (comp['value']*units(comp['units'])*sample_mass).to('mg')
                
            else:
                #assume concs for now...
                if len(next_sample_dict)<(len(self.components)-1):
                    raise ValueError('System under specified...')

                mass_dict = {}
                for name,comp in next_sample_dict.items():
                    mass_dict[name] = (comp['value']*units(comp['units'])*sample_volume).to('mg')
            
            self.target = AFL.automation.prepare.Solution('target',self.components)
            for k,v in mass_dict.items():
                self.target[k].mass = v
            self.sample = Sample('dummy',self.target,target_check=self.target,balancer=None)
            validated=True

            #######################################
            ## INIITLIAZE DATASET FOR NEW SAMPLE ##
            #######################################
            sample_uuid = str(uuid.uuid4())
            sample_name = f'AL_{self.config["data_tag"]}_{sample_uuid[-8:]}'

            self.new_data              = xr.Dataset()
            self.new_data['sample_name']     = sample_name
            self.new_data['label']     = -1
            self.new_data['validated'] = validated
            self.new_data['sample_uuid']      = sample_uuid
            sample_composition = {}
            if ternary:
                total = 0
                for component in self.AL_components:
                    mf = self.sample.target_check.mass_fraction[component].magnitude
                    self.new_data[component] = mf
                    total+=mf
                for component in self.AL_components:
                    self.new_data[component] = self.new_data[component]/total
                    
                    sample_composition[component] = {}
                    sample_composition[component]['value'] = self.new_data[component].values[()]
                    sample_composition[component]['units'] = 'fraction'
            else:
                for component in self.AL_components:
                    self.new_data[component] = self.sample.target_check.concentration[component].to("mg/ml").magnitude
                    self.new_data[component].attrs['units'] = 'mg/ml'
                    
                    sample_composition[component] = {}
                    sample_composition[component]['value'] = self.new_data[component].values[0]
                    sample_composition[component]['units'] = 'mg/ml'
                    
                
            for component in self.components:
                self.new_data['mfrac_'+component] = self.sample.target_check.mass_fraction[component].magnitude
                self.new_data['mass_'+component] = self.sample.target_check[component].mass.to('mg').magnitude
                self.new_data['mass_'+component].attrs['units'] = 'mg'
                    
            
            #################
            ##  SET SAMPLE ##
            #################
            sample_data = self.set_sample(sample_name=sample_name,sample_uuid=sample_uuid, sample_composition=sample_composition)
            print(sample_data)
            #self.prep_client.enqueue(task_name='set_sample',**sample_data)
            #self.load_client.enqueue(task_name='set_sample',**sample_data)
            self.sas_client.enqueue(task_name='set_sample',**sample_data)
            self.agent_client.enqueue(task_name='set_sample',**sample_data)
            # if self.spec_client is not None:
            #     self.spec_client.enqueue(task_name='set_sample',**sample_data)
                
            ####################
            ## PROCESS SAMPLE ##
            ####################
            #self.process_sample(
            self.measure(
                    dict(
                        name=sample_name,
                        prep_protocol = [],
                        catch_protocol =[],
                        volume = 1.0,
                        exposure = 1.0,
                        empty_exposure = 0.0
                )
            )
            
            ########################
            ## PROCESS MESUREMENT ##
            ########################
            res = self.tiled_cat.search(Eq('sample_uuid',sample_uuid)).search(Eq('task_name','expose'))
            if len(res)==0:
                self.app.logger.info(f'No expose task found for {sample_uuid}')
                self.AL_status_str = 'No expose task found'
                raise ValueError('No expose task found')
            array_client = list(res.items())[-1][-1]#first [-1] is for last expose, second is for actual ArrayClient rather than key
            self.new_data['SAS'] = xr.DataArray(
                data=array_client.metadata['I'],
                dims=['q'],
                coords={'q':array_client.metadata['q']}
            )
            
            ########################
            ## UPDATE AL MANIFEST ##
            ########################
            if start_new_manifest and AL_manifest_path.exists():
                self.update_status(f'Backing up old and then starting new manifest...')
                AL_manifest_backup = xr.load_dataset(AL_manifest_path)
                i=0
                while True:
                    AL_manifest_backup_path = pathlib.Path(str(AL_manifest_path)+f'.bak{i}')
                    if AL_manifest_backup_path.exists():
                        i+=1
                    else:
                        print(f'--> Writing backup AL file to {AL_manifest_backup_path}')
                        AL_manifest_backup.to_netcdf(AL_manifest_backup_path)
                        break

                self.AL_manifest = self.new_data.expand_dims('sample')
                start_new_manifest=False

            elif AL_manifest_path.exists():
                self.update_status(f'Updating manifest...')
                self.AL_manifest = xr.load_dataset(AL_manifest_path)

                if 'grid' in self.AL_manifest.dims:
                    self.AL_manifest = self.AL_manifest.drop_dims('grid')

                self.AL_manifest = xr.concat([self.AL_manifest,self.new_data],dim='sample')

            else:
                self.update_status(f'Starting new manifest...')
                self.AL_manifest = self.new_data.expand_dims('sample')
            
            self.AL_manifest.attrs.update(self.AL_kwargs)
            self.AL_manifest.attrs['AL_data'] = 'SAS'
            self.AL_manifest.attrs['components'] = self.AL_components

            #self.AL_manifest  = self.AL_manifest.afl.comp.add_grid(pts_per_row=pts_per_row)

            self.AL_manifest.to_netcdf(AL_manifest_path)

            ################
            ## TRIGGER AL ##
            ################
            self.update_status(f'Triggering agent server...')
            if len(pre_run_list)==0:
                if predict:
                    self.agent_uuid = self.agent_client.enqueue(task_name='predict',datatype='nc')
                
                    # wait for AL
                    self.app.logger.info(f'Waiting for agent...')
                    self.agent_client.wait(self.agent_uuid)
                else:#used for intialization
                    return
            
            






   
